#!/usr/bin/env python
"""
Generate the images and rst files for gallery of SfePy examples.

The following steps need to be made to regenerate the documentation with the
updated example files:

1. remove doc/examples/*::

   $ rm -rf doc/examples/*

2. generate the files:

   $ ./tools/gen_gallery.py

3. regenerate the documentation::

   $ python setup.py htmldocs
"""
import sys
sys.path.append('.')
import os
import tempfile
import glob
import re
import shutil
import subprocess
from itertools import chain
from argparse import ArgumentParser, RawDescriptionHelpFormatter

import numpy as nm
import matplotlib.image as image

import sfepy
from sfepy.base.base import (get_default, ordered_iteritems,
                             import_file, output, Struct)
from sfepy.base.ioutils import (ensure_path, locate_files, remove_files,
                                edit_filename)
from sfepy.scripts.resview import pv_plot, get_camera_position
import pyvista as pv

omits = [
    '__init__.py',
]

omit_images = [
    'band_gaps_conf.py',
    'dg_plot_1D.py',
    'example_dg_common.py',
    'homogenization_opt.py', # Is parameterized.
    'linear_elasticity_opt.py', # Is parameterized.
    'linear_homogenization_postproc.py',
    'linear_homogenization_up.py', # Used in linear_elastic_mM.py.
    'material_opt.py', # Very long calculation.
    'nonlinear_homogenization.py',
    'piezo_elasticity_micro.py', # Used in piezo_elasticity_macro.py.
    'quantum_common.py',
]

omit_dirs = [
    re.compile('.*output.*/').match,
]

custom = {
    'acoustics/vibro_acoustic3d.py': {
        '_Gamma0': {'view_2d': True, 'max_plots': 2},
        '_Omega1': {'camera': [45, 55, 0.8]},
        '_Omega2': {'camera': [45, 55, 0.8]},
    },
    'acoustics/acoustics3d.py': {
        '_Omega_1': {
            'camera': [75, 135, 1.4],
            'grid_vector1': [1.2, 0, 0],
        },
        '_Omega_2': {
            'camera': [75, 135, 1],
            'grid_vector1': [1.2, 0, 0],
        },
    },
    'acoustics/helmholtz_apartment.py': {
        '': {
            'fields': ['imag.E:wimag.E:f10%:p0', 'mat_id:p1'],
            'force_view_3d': True,
            'grid_vector1': [1.1, 0, 0],
            'camera_position': [-9.22684,-8.37688,10.7623,
                                1.51653,0.122742,-0.646791,
                                0.472758,0.432784,0.767593],
            'color_map': 'seismic',
        },
    },
    'diffusion/cube.py': {
        '': {'camera': [225, 55, 0.7]},
    },
    'diffusion/laplace_time_ebcs.py': {
        '': {'camera': [225, 55, 0.7]},
    },
    'diffusion/laplace_fluid_2d.py': {
        '': {'fields': ['phi:p0', 'phi:t50:p0']},
    },
    'diffusion/laplace_1d.py': {
        '': {
            'fields': ['t:wt:p0', 't:p0'],
            'force_view_3d': True,
            'camera_position': [0.243787,-1.51098,0.486608,
                                0.5,0,0,
                                0,0,1],
        },
    },
    'dg/advection_1D.py': {
        '': {
            'fields': ['p_modal0:r:wp_modal0:p0', '1:vw:p0',
                       'p_modal1:r:wp_modal1:p1', '1:vw:p1',
                       'p_modal2:r:wp_modal2:p2', '1:vw:p2'],
            'force_view_3d': True,
            'camera_position': [0.300893,-2.41044,0.408743,
                                0.4,0,0.55728,
                                0,0,1],
            'grid_vector2': [0, 0, 55],
            'max_plots': 1,
            'color_map': 'cool',
        },
    },
    'diffusion/laplace_coupling_lcbcs.py': {
        '': {
            'fields': ['u1:wu1:p0', 'u1:vw:p0',
                       'u2:wu2:p1', 'u2:vw:p1'],
            'force_view_3d': True,
        },
    },
    'diffusion/laplace_iga_interactive.py': {
        'command': 'python3 sfepy/examples/diffusion/laplace_iga_interactive.py -o output',
        'result': 'output/concentric_circles.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'diffusion/laplace_refine_interactive.py': {
        'command': 'python3 sfepy/examples/diffusion/laplace_refine_interactive.py output',
        'result': 'output/hanging.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['u:wu', '1:vw'],
                'camera_position': [-1.27046,-1.02066,2.28476,
                                    -0.0525852,0.116151,0.00474463,
                                    0.169164,0.842974,0.510664],
            },
        },
    },
    'diffusion/laplace_shifted_periodic.py': {
        'command': 'python3 sfepy/examples/diffusion/laplace_shifted_periodic.py',
        'result': 'laplace_shifted_periodic.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['u:wu:f0.5', '1:vw'],
                'camera_position': [-1.42703,0.70237,1.33953,
                                    -0.0817071,-0.0547046,0.0280939,
                                    0.616241,-0.216582,0.757192],
            },
        },
    },
    'diffusion/poisson_iga.py': {
        '': {
            'fields': ['t:wt:p0', 't:vw:o0.4:p0'],
            'force_view_3d': True,
            'camera': [30, 60, 1.],
        },
    },
    'diffusion/poisson_parallel_interactive.py': {
        'command': 'python3 sfepy/examples/diffusion/poisson_parallel_interactive.py output-parallel -2 --shape=101,101',
        'result': 'output-parallel/sol.h5',
        'sfepy-view-options': {
            '': {
                'fields': ['u:wu', '1:vw'],
                'camera_position': [-1.00607,-1.49437,0.843106,
                                    0.0141441,-0.0501477,-0.0173962,
                                    0.309532,0.315757,0.896932],
            },
        },
    },
    'diffusion/poisson_parametric_study.py': {
        'command': 'sfepy-run sfepy/examples/diffusion/poisson_parametric_study.py',
        'result': 'output/r_omega1/circles_in_square_0_1_2_3_4_5_6.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'diffusion/sinbc.py': {
        '_t': {
            'fields': ['t:wt:p0', '1:vw:p0'],
            'force_view_3d': True,
            'camera': [0, -45, 1.],
        },
        '_grad': {
            'fields': ['grad:g:f0.01:p0', '1:vw:p0'],
            'force_view_3d': True,
            'camera': [0, -45, 1.5],
        },
    },
    'diffusion/time_heat_equation_multi_material.py': {
        '': {
            'isosurfaces': 10,
            'outline': True,
            'camera': [-50, -230, 1],
        },
    },
    'diffusion/time_poisson_interactive.py': {
        'command': 'python3 sfepy/examples/diffusion/time_poisson_interactive.py -p',
        'image': 'time_poisson_interactive_probe_04.png',
        'result': 'domain.10.vtk',
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'homogenization/linear_homogenization.py': {
        'command': 'sfepy-run sfepy/examples/homogenization/linear_homogenization.py',
        'result': 'output/corrs_le.vtk',
        'sfepy-view-options': {
            '': {
                'max_plots': 3,
                'camera_position': [-8.57147,2.10122,2.76537,
                                    0.37,2.04405,2.14477,
                                    0,0,1],
                'show_labels': False,
                'show_scalar_bars': False,
            }
        },
    },
    'homogenization/perfusion_micro.py': {
        'command': 'sfepy-run sfepy/examples/homogenization/perfusion_micro.py',
        'result': 'output/corrs_3d_2ch.vtk',
        'sfepy-view-options': {
            '_etaA_YM': {
                'camera_position': [-1.69531,-0.966016,2.3648,
                                    0.480252,0.486045,0.46669,
                                    0.51024,0.292658,0.808707],
            },
            '_etaB_YM': {
                'camera_position': [-1.69531,-0.966016,2.3648,
                                    0.480252,0.486045,0.46669,
                                    0.51024,0.292658,0.808707],
            },
            # '_gamma_A_1_YA': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_A_2_YA': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_A_3_YA': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_B_1_YB': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_B_2_YB': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_B_3_YB': {
            #     'camera_position': [-0.96887,-2.21765,1.25047,
            #                         1,0.4,0.5,
            #                         0.107002,0.198813,0.974179],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_gamma_m_YM': {
            #     'camera_position': [-1.69531,-0.966016,2.3648,
            #                         0.480252,0.486045,0.46669,
            #                         0.51024,0.292658,0.808707],
            # },
            '_gamma_p_YM': {
                'camera_position': [-1.69531,-0.966016,2.3648,
                                    0.480252,0.486045,0.46669,
                                    0.51024,0.292658,0.808707],
            },
            # '_oneA_YM': {
            #     'camera_position': [-1.69531,-0.966016,2.3648,
            #                         0.480252,0.486045,0.46669,
            #                         0.51024,0.292658,0.808707],
            # },
            # '_oneB_YM': {
            #     'camera_position': [-1.69531,-0.966016,2.3648,
            #                         0.480252,0.486045,0.46669,
            #                         0.51024,0.292658,0.808707],
            # },
            # '_piA_YA': {
            #     'max_plots': 2,
            #     'camera_position': [-2.02509,0.148141,3.9141,
            #                         0.913524,1.35355,0.462869,
            #                         0.714983,0.186315,0.673859],
            #     'grid_vector1': [1.2,0,0],
            # },
            # '_piB_YB': {
            #     'max_plots': 2,
            #     'camera_position': [-2.02509,0.148141,3.9141,
            #                         0.913524,1.35355,0.462869,
            #                         0.714983,0.186315,0.673859],
            #     'grid_vector1': [1.2,0,0],
            # },
        },
    },
    'homogenization/rs_correctors.py': {
        'command': 'python3 sfepy/examples/homogenization/rs_correctors.py -n',
        'result': 'corrs_elastic.vtk',
        'sfepy-view-options': {
            '': {
                'max_plots': 2,
                'camera_position': [1.16983,0.660181,7.63496,
                                    1.16983,0.660181,
                                    0,0,1,0],
            },
        },
    },
    'large_deformation/compare_elastic_materials.py': {
        'command': 'python3 sfepy/examples/large_deformation/compare_elastic_materials.py -n',
        'image': 'pressure_displacement.png',
        'sfepy-view-options': {
        },
    },
    'large_deformation/gen_yeoh_tl_up_interactive.py': {
        'command': 'python3 sfepy/examples/large_deformation/gen_yeoh_tl_up_interactive.py -pn',
        'image': 'gen_yeoh_tl_up_comparison.png',
        'result': 'domain.10.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['stress:wu:f1:p0', '1:vw:p0'],
                'camera_position': [-1.89068,-2.47529,0.674151,
                                    0.793688,0.209075,0.574171,
                                    0,0,1],
            }
        },
    },
    'large_deformation/hyperelastic_tl_up_interactive.py': {
        'command': 'python3 sfepy/examples/large_deformation/hyperelastic_tl_up_interactive.py -pn',
        'image': 'hyperelastic_tl_up_comparison.png',
        'result': 'domain.10.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['stress:wu:f1:p0', '1:vw:p0'],
                'camera_position': [-4.04955,-6.37091,3.78116,
                                    2.06774,-0.866196,1.16309,
                                    0.220544,0.208124,0.952914],
            }
        },
    },
    'linear_elasticity/dispersion_analysis.py': {
        'command_0': 'python3 sfepy/examples/linear_elasticity/dispersion_analysis.py meshes/2d/special/circle_in_square.mesh --log-std-waves --eigs-only --no-show',
        'command_1': 'python3 sfepy/scripts/plot_logs.py output/frequencies.txt -o output/frequencies.png -n',
        'image': 'output/frequencies.png',
        'sfepy-view-options': {
        },
    },
    'linear_elasticity/elastic_contact_planes.py': {
        '': {
            'fields': ['u:wu:p0', '1:vw:p0'],
            'camera': [225, 55, 0.7],
        },
    },
    'linear_elasticity/elastic_contact_sphere.py': {
        '': {
            'fields': ['u:wu:p0', '1:vw:p0'],
            'camera': [225, 55, 0.7],
        },
    },
    'linear_elasticity/elastic_shifted_periodic.py': {
        '': {'fields': ['von_mises_stress:r:wu:p0', '1:vw:p0']},
    },
    'linear_elasticity/elastodynamic.py': {
        '': {
            'fields': ['u:wu:f1e3:p0', '1:vw:p0',
                       'cauchy_strain:p1', 'cauchy_stress:p2'],
        },
    },
    'linear_elasticity/elastodynamic_identification.py': {
        'command_0': 'python3 sfepy/examples/linear_elasticity/elastodynamic_identification.py --output-dir=output/edi',
        'image_0': 'output/edi/res00004.png',
        'command_1': 'python3 sfepy/scripts/plot_logs.py output/edi/pars.txt -o output/edi/pars.png -n',
        'image_1': 'output/edi/pars.png',
        'sfepy-view-options': {
        },
    },
    'linear_elasticity/its2D_4.py': {
        'command_0': 'sfepy-run sfepy/examples/linear_elasticity/its2D_4.py',
        'command_1': 'sfepy-probe sfepy/examples/linear_elasticity/its2D_4.py its2D.h5',
        'image_0': 'its2D_0.png',
        'image_1': 'its2D_1.png',
        'result': 'its2D.h5',
        'dim': 2,
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'linear_elasticity/its2D_5.py': {
        'command': 'sfepy-run sfepy/examples/linear_elasticity/its2D_5.py',
        'image_0': 'its2D_probe_line0.png',
        'image_1': 'its2D_probe_line1.png',
        'result': 'its2D.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'linear_elasticity/its2D_interactive.py': {
        'command': 'python3 sfepy/examples/linear_elasticity/its2D_interactive.py -p',
        'image_0': 'its2D_interactive_probe_0.png',
        'image_1': 'its2D_interactive_probe_1.png',
        'result': 'its2D_interactive.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '': {
            },
        },
    },
    'linear_elasticity/linear_elastic_iga.py': {
        '': {
            'fields': ['u:wu:p0', '1:vw:p0'],
            'camera': [-45, 55, 1],
        },
    },
    'linear_elasticity/linear_elastic_interactive.py': {
        'command': 'python3 sfepy/examples/linear_elasticity/linear_elastic_interactive.py',
        'result': 'linear_elasticity.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '': {
                'fields': ['u:wu:p0', '1:vw:p0'],
            },
        },
    },
    'linear_elasticity/linear_elastic_probes.py': {
        'command': 'sfepy-run sfepy/examples/linear_elasticity/linear_elastic_probes.py',
        'image_0': 'cylinder_probe_line.png',
        'image_1': 'cylinder_probe_circle.png',
        'image_2': 'cylinder_probe_ray.png',
        'result': 'cylinder.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['cauchy_stress:wu:p0', '1:vw:p0'],
                'camera_position': [-0.0496419,-0.132205,0.0339214,
                                    0.0495001,0.00309172,-0.00112878,
                                    0.146854,0.145842,0.978348],
            },
        },
    },
    'linear_elasticity/linear_viscoelastic.py': {
        '': {'camera': [225, 75, 0.88]}
    },
    'linear_elasticity/modal_analysis.py': {
        'command': 'python3 sfepy/examples/linear_elasticity/modal_analysis.py  -s 31,31 -n 3',
        'result': 'eigenshapes.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['strain000:wu000:f10%:p0', '1:vw:p0',
                           'strain001:wu001:f10%:p1', '1:vw:p1',
                           'strain002:wu002:f10%:p2', '1:vw:p2'],
                'camera_position': [1.1,-0.1,5.41561,
                                    1.1,-0.1,
                                    0,0,1,0],
                'grid_vector1': [1.2,0,0],
            },
        },
    },
    'linear_elasticity/modal_analysis_declarative.py': {
        '': {
            'fields': ['u003:wu003:f30%:p0', '1:vw:p0'],
            'camera_position': [-2.30562,-2.2604,-0.325838,
                                -0.0714771,0.0374911,0.0287214,
                                -0.0245554,-0.129086,0.991329],
        },
    },
    'linear_elasticity/multi_node_lcbcs.py': {
        '': {
            'fields': ['u:wu:e'],
            'force_view_3d': True,
            'camera_position': [0.175,0.125,0.735014,
                                0.175,0.125,0,
                                0,1,0],
        },
    },
    'linear_elasticity/seismic_load.py': {
        '': {
            'fields': ['cauchy_stress:wu:f10:p0', '1:vw:p0'],
        },
    },
    'linear_elasticity/shell10x_cantilever.py': {
        '': {
            'fields': ['u_disp:wu_disp:p0', '1:vw:p0',
                       'u_rot:p1', '1:vw:p1'],
            'camera': [-45, 75, 1],
            'grid_vector1': [1, 0, 0],
        },
    },
    'linear_elasticity/shell10x_cantilever_interactive.py': {
        'command': 'python3 sfepy/examples/linear_elasticity/shell10x_cantilever_interactive.py -t bend --plot --no-show output',
        'image': 'output/shell10x_cantilever_convergence_bent.png',
        'result': 'output/shell10x_cantilever.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['u_disp:wu_disp:p0', '1:vw:p0',
                           'u_rot:p1', '1:vw:p1'],
                'camera': [-45, 75, 1],
                'grid_vector1': [1, 0, 0],
            },
        },
    },
    'linear_elasticity/truss_bridge.py': {
        '': {
            'view_2d': True,
            'fields': ['u:wu:p0', '1:vw:p0', 'S:e:p1'],
            'grid_vector1': [0, 2, 0],
        },
    },
    'linear_elasticity/truss_bridge3d.py': {
        '_Solid': {
            'fields': ['u_solid:wu_solid:f1e3:p0'],
            'camera_position': [-5.912, -6.64883, 1.80888,
                                5.05199, 2.28013, -1.49468,
                                0, 0, 1],
        },
        '_Struct': {
            'fields': ['u_struct:wu_struct:f1e3:p0'],
            'camera_position': [-5.912, -6.64883, 1.80888,
                                5.05199, 2.28013, -1.49468,
                                0, 0, 1],
        },
    },
    'linear_elasticity/wedge_mesh.py': {
        'command': 'sfepy-run sfepy/examples/linear_elasticity/wedge_mesh.py',
        'result': 'beam_w14.vtk',
        'sfepy-view-options': {
            '': {
                'fields': ['u:wu:e:o0.5'],
                'camera_position': [0.927482,-0.574865,0.307926,
                                    0.372897,0.120369,-0.0347131,
                                    -0.326236,0.19556,0.924838],
            }
        },
    },
    'miscellaneous/live_plot.py': {
        'command_0': 'python3 sfepy/examples/miscellaneous/live_plot.py -o output',
        'command_1': 'python3 sfepy/scripts/plot_logs.py output/live_plot.txt -o output/live_plot.png -n',
        'image_1': 'output/live_plot.png',
        'command_2': 'python3 sfepy/scripts/plot_logs.py output/live_plot2.txt -o output/live_plot2.png -n',
        'image_2': 'output/live_plot2.png',
        'sfepy-view-options': {
        },
    },
    'miscellaneous/refine_evp.py': {
        'command': 'python3 sfepy/examples/miscellaneous/refine_evp.py --max-order=5 --max-refine=2 --fig-suffix=.png --no-show',
        'image': 'output/h-refinement-0-laplace-lagrange-primme-none-a.png',
        'sfepy-view-options': {
        },
    },
    'multi_physics/biot_parallel_interactive.py': {
        'command': 'python3 sfepy/examples/multi_physics/biot_parallel_interactive.py output-parallel',
        'result': 'output-parallel/sol.h5',
        'sfepy-view-options': {
            '': {
                'fields': ['u:t1000:p0', '1:vw:o0.1:p0', 'p:p1'],
                'camera_position': [-0.740809,-2.67483,2.43604,
                                    0.579298,-0.106475,0.0404841,
                                    0.183213,0.618527,0.764106],
                'grid_vector1': [1.2, 0, 0]
            },
        },
    },
    'multi_physics/piezo_elasticity.py': {
        '': {'fields': ['u:g:p0', 'cauchy_strain:p1',
                        'elastic_stress:p2', 'piezo_stress:p3'],
             'grid_vector1': [1.2, 0, 0],
             'max_plots': 4},
    },
    'multi_physics/piezo_elastodynamic.py': {
        '': {
            'fields': ['p:wu:f2000:p0', '1:vw:wu:f2000:p0'],
            'camera_position': [-0.0125588,-0.00559266,0.0117482,
                                0.00438669,0.00487109,0.00135715,
                                0.334487,0.333581,0.881387],
            'color_map': 'inferno',
        }
    },
    'multi_physics/thermal_electric.py': {
        'command': 'python3 sfepy/examples/multi_physics/thermal_electric.py',
        'result': 'sfepy/examples/multi_physics/output/circle_in_square.vtk',
        'dim': 2,
        'sfepy-view-options': {
            '.el': {
            },
            '.10': {
            },
        },
    },
    'multi_physics/thermo_elasticity_ess.py': {
        '': {
            'fields': ['T:wu:f1e3:p0', '1:vw:p0'],
            'camera': [-45, 75, 1],
        },
    },
    'multi_physics/thermo_elasticity.py': {
        '': {'camera': [225, 75, 0.9]}
    },
    'navier_stokes/stokes_slip_bc.py': {
        '': {
            'fields': ['u:g:f.25:p0', 'u:o.4:p0', 'p:p1'],
            'camera': [-45, 55, 1],
            'grid_vector1': [0, 1.2, 0]
        },
    },
    'phononic/band_gaps.py': {
        'command': 'sfepy-run sfepy/examples/phononic/band_gaps.py --phonon-band-gaps --phonon-plot -O plot_options={show=False,legend=True},fig_suffix=\'.png\'',
        'image': 'sfepy/examples/phononic/output/band_gaps/band_gaps.png',
        'result': 'sfepy/examples/phononic/output/band_gaps/evp.vtk',
        'sfepy-view-options': {
            '': {
                'camera_position': [1.8,-1.2,7.28269,
                                    1.8,-1.2,0,
                                    0,1,0],
                'grid_vector1': [1.2, 0, 0],
                'grid_vector2': [0, -1.2, 0],
                'max_plots': 4,
                'show_labels': False,
                'show_scalar_bars': False,
            },
        },
    },
    'phononic/band_gaps_rigid.py': {
        'command': 'sfepy-run sfepy/examples/phononic/band_gaps_rigid.py --phonon-band-gaps --phonon-plot -O plot_options={show=False,legend=True},fig_suffix=\'_rigid.png\'',
        'image': 'sfepy/examples/phononic/output/band_gaps_rigid/band_gaps_rigid.png',
        'result': 'sfepy/examples/phononic/output/band_gaps_rigid/evp.vtk',
        'sfepy-view-options': {
            '': {
                'fields': list(chain(
                    *[[f'u{ii:03d}:vs:o.4:p{ii}', f'u{ii:03d}:g:p{ii}']
                      for ii in range(12)]
                )),
                'camera_position': [1.8,-1.2,7.28269,
                                    1.8,-1.2,0,
                                    0,1,0],
                'grid_vector1': [1.2, 0, 0],
                'grid_vector2': [0, -1.2, 0],
                'max_plots': 4,
                'show_labels': False,
                'show_scalar_bars': False,
            },
        },
    },
    'quantum/boron.py': {
        '': {'fields': ['Psi000:p0', 'Psi001:p1', 'Psi002:p2', 'Psi003:p3'],
             'grid_vector1': [1.2, 0, 0],
             'max_plots': 4},
    },
    'quantum/hydrogen.py': {
        '': {'fields': ['Psi000:p0', 'Psi001:p1', 'Psi002:p2', 'Psi003:p3'],
             'grid_vector1': [1.2, 0, 0],
             'max_plots': 4},
    },
    'quantum/oscillator.py': {
        '': {'fields': ['Psi000:p0', 'Psi001:p1', 'Psi002:p2', 'Psi003:p3'],
             'grid_vector1': [1.2, 0, 0],
             'max_plots': 4},
    },
    'quantum/well.py': {
        '': {'fields': ['Psi000:p0', 'Psi001:p1', 'Psi002:p2', 'Psi003:p3'],
             'grid_vector1': [1.2, 0, 0],
             'max_plots': 4},
    },
}


def resview_plot(filename, filename_out, options):
    pv.set_plot_theme("document")
    plotter = pv.Plotter(off_screen=True)

    plotter = pv_plot([filename], options, plotter=plotter, use_cache=False)
    if options.axes_visibility:
        plotter.add_axes(**dict(options.axes_options))

    if options.view_2d:
        plotter.view_xy()
        plotter.show(screenshot=filename_out, window_size=(800, 600))
    else:
        if options.camera_position is not None:
            cpos = nm.array(options.camera_position)
            cpos = cpos.reshape((3, 3))
        elif options.camera:
            zoom = options.camera[2] if len(options.camera) > 2 else 1.
            cpos = get_camera_position(plotter.bounds,
                                       options.camera[0], options.camera[1],
                                       zoom=zoom)
        else:
            cpos = None

        plotter.show(cpos=cpos, screenshot=filename_out, window_size=(800, 600))

def run_resview_plot(*args):
    """
    A fix for the problem that calling :func:`resview_plot()` directly often
    terminates the program.
    """
    from multiprocessing import Process

    process = Process(target=resview_plot, args=args)
    process.start()
    process.join()

def _omit(filename, omits, omit_dirs):
    omit = False

    base = os.path.basename(filename)

    if base in omits:
        omit = True

    for omit_dir in omit_dirs:
        if omit_dir(filename) is not None:
            omit = True
            break

    return omit


def ebase2fbase(ebase):
    return os.path.splitext(ebase)[0].replace(os.path.sep, '-')

def _make_fig_name(fig_base, fig_filename):
    return fig_base + '-' + os.path.basename(fig_filename)

def _get_image_names(custom_options):
    for key, val in custom_options.items():
        if key.startswith('image'):
                yield val

def _get_fig_filenames(ebase, images_dir):
    fig_base = ebase2fbase(ebase)

    if ebase in custom:
        custom_options = custom.get(ebase)
        if 'sfepy-view-options' in custom_options:
            custom_view_options = custom_options['sfepy-view-options']

            for fig_filename in _get_image_names(custom_options):
                yield os.path.join(images_dir,
                                   _make_fig_name(fig_base, fig_filename))

        else:
            custom_view_options = custom_options

        if custom_view_options:
            suffixes = sorted(custom_view_options.keys())
            for suffix in suffixes:
                yield os.path.join(images_dir, fig_base + suffix + '.png')

    else:
        yield os.path.join(images_dir, fig_base + '.png')


def _get_fig_filename(ebase, images_dir, suffix):
    fig_base = ebase2fbase(ebase)

    return os.path.join(images_dir, fig_base + suffix + '.png')


def _make_sphinx_path(path, relative=False):
    if relative:
        aux = path.replace(sfepy.data_dir, '')
        prefix = ('..' + os.path.sep) * aux.count(os.path.sep)
        sphinx_path = prefix[:-1] + aux

    else:
        sphinx_path = path.replace(sfepy.data_dir, '/..')

    return sphinx_path


def _apply_commands(custom_options, ebase, images_dir):
    for key, val in custom_options.items():
        if key.startswith('command'):
            cmd = custom_options[key]
            subprocess.run(cmd.split()).check_returncode()

        if key.startswith('image'):
            shutil.copy(val,
                        os.path.join(images_dir,
                                     _make_fig_name(ebase2fbase(ebase), val)))


def apply_view_options(views, default):
    out = {}
    for kview, view in views.items():
        ov = default.copy()
        for k, v in view.items():
            if k == 'fields':
                fv = [k.split(':') for k in v]
                fv = [(k[0], ':'.join(k[1:])) for k in fv]
                setattr(ov, k, fv)
            else:
                setattr(ov, k, v)

        out[kview] = ov

    return out


def generate_images(images_dir, examples_dir, pattern='*.py'):
    """
    Generate images from results of running examples found in
    `examples_dir` directory.

    The generated images are stored to `images_dir`,
    """
    from sfepy.applications import solve_pde
    from sfepy.solvers.ts_solvers import StationarySolver

    prefix = output.prefix

    output_dir = tempfile.mkdtemp()
    trunk = os.path.join(output_dir, 'result')
    options = Struct(output_filename_trunk=trunk,
                     output_format='vtk',
                     save_ebc=False,
                     save_ebc_nodes=False,
                     save_regions=False,
                     save_regions_as_groups=False,
                     solve_not=False)

    view_options = Struct(step=0,
                          fields=[], fields_map=[],
                          outline=False,
                          isosurfaces=0,
                          show_edges=False,
                          warp=None,
                          factor=1.,
                          opacity=1.,
                          color_map='viridis',
                          axes_options=[],
                          axes_visibility=False,
                          grid_vector1=None,
                          grid_vector2=None,
                          max_plots=3,
                          show_labels=False,
                          label_position=[-1, -1, 0, 0.2],
                          scalar_bar_size=[0.15, 0.06],
                          scalar_bar_position=[0.04, 0.92, 0, -1.5],
                          show_scalar_bars=True,
                          camera=[225, 75, 1],
                          camera_position=None,
                          view_2d=False,
                          force_view_3d=False,
                          show_step_time=False)

    ensure_path(images_dir + os.path.sep)

    for ex_filename in locate_files(pattern, examples_dir):
        if _omit(ex_filename, omits + omit_images, omit_dirs):
            continue

        output.level = 0
        output.prefix = prefix
        ebase = ex_filename.replace(examples_dir, '')[1:]
        output('trying "%s"...' % ebase)

        _ebase = ebase.replace(os.path.sep, '/')
        custom_options = custom.get(_ebase)
        if custom_options and 'sfepy-view-options' in custom_options:
            try:
                _apply_commands(custom_options, ebase, images_dir)

                filename = custom_options.get('result')
                dim = custom_options.get('dim')
                custom_view_options = custom_options['sfepy-view-options']

            except subprocess.CalledProcessError:
                filename = None
                output('***** failed! *****')

        else:
            custom_view_options = custom_options

            try:
                problem, state = solve_pde(ex_filename, options=options)
                try:
                    tsolver = problem.get_solver()

                except ValueError:
                    suffix = None

                else:
                    if isinstance(tsolver, StationarySolver):
                        suffix = None

                    else:
                        suffix = tsolver.ts.suffix % (tsolver.ts.n_step - 1)

                filename = problem.get_output_name(suffix=suffix)
                dim = problem.get_dim()

            except KeyboardInterrupt:
                raise

            except:
                filename = None
                output('***** failed! *****')

        if filename is not None:
            if custom_view_options is not None:
                views = apply_view_options(custom_view_options, view_options)
            else:
                views = {'': view_options.copy()}

            for suffix, kwargs in views.items():
                if dim in (1, 2) and not kwargs.force_view_3d:
                    kwargs.view_2d = True
                    kwargs.scalar_bar_position = [0.04, 0.92, 1.7, 0]
                    if kwargs.grid_vector1 is None:
                        kwargs.grid_vector1 = [1.2, 0, 0]

                    if kwargs.grid_vector2 is None:
                        kwargs.grid_vector2 = [0, -1.2, 0]

                fig_filename = _get_fig_filename(ebase, images_dir, suffix)

                fname = edit_filename(filename, suffix=suffix)
                output('displaying results from "%s"' % fname)
                disp_name = fig_filename.replace(sfepy.data_dir, '')
                output('to "%s"...' % disp_name.lstrip(os.path.sep))

                run_resview_plot(fname, fig_filename, kwargs)

                output('...done')

            remove_files(output_dir)

        output('...done')


def generate_thumbnails(thumbnails_dir, images_dir, scale=0.3):
    """
    Generate thumbnails into `thumbnails_dir` corresponding to images in
    `images_dir`.
    """
    ensure_path(thumbnails_dir + os.path.sep)

    output('generating thumbnails...')
    filenames = glob.glob(os.path.join(images_dir, '*.png'))
    for fig_filename in filenames:
        ebase = fig_filename.replace(sfepy.data_dir, '').lstrip(os.path.sep)
        output('"%s"' % ebase)

        base = os.path.basename(fig_filename)
        thumb_filename = os.path.join(thumbnails_dir, base)

        image.thumbnail(fig_filename, thumb_filename, scale=scale)

    output('...done')


_index = """\
.. _%s-index:

%s
%s

.. toctree::
   :maxdepth: 2

"""

_image = '.. image:: %s'

_include = """\
.. _%s:

%s
%s

**Description**

%s

%s

:download:`source code <%s>`

.. literalinclude:: %s

"""


def generate_rst_files(rst_dir, examples_dir, images_dir, pattern='*.py'):
    """
    Generate Sphinx rst files for examples in `examples_dir` with images
    in `images_dir` and put them into `rst_dir`.

    Returns
    -------
    dir_map : dict
        The directory mapping of examples and corresponding rst files.
    """
    ensure_path(rst_dir + os.path.sep)

    output('generating rst files...')

    dir_map = {}
    for ex_filename in locate_files(pattern, examples_dir):
        if _omit(ex_filename, omits, omit_dirs):
            continue

        ebase = ex_filename.replace(examples_dir, '')[1:]
        base_dir = os.path.dirname(ebase)
        rst_filename = ebase2fbase(ebase) + '.rst'
        dir_map.setdefault(base_dir, []).append((ex_filename, rst_filename))

    for dirname, filenames in dir_map.items():
        filenames = sorted(filenames, key=lambda a: a[1])
        dir_map[dirname] = filenames

    # Main index.
    mfd = open(os.path.join(rst_dir, 'index.rst'), 'w')
    mfd.write(_index % ('examples', 'Examples', '=' * 8))

    for dirname, filenames in ordered_iteritems(dir_map):
        # Subdirectory index.
        ifd = open(os.path.join(rst_dir, '%s-index.rst' % dirname), 'w')
        ifd.write(_index % (dirname + '-examples',
                            dirname, '=' * len(dirname)))

        for ex_filename, rst_filename in filenames:
            full_rst_filename = os.path.join(rst_dir, rst_filename)
            output('"%s"' % rst_filename)
            ebase = ex_filename.replace(examples_dir, '')[1:]

            rst_base = os.path.splitext(rst_filename)[0]

            rst_ex_filename = _make_sphinx_path(ex_filename)
            try:
                docstring = get_default(import_file(ex_filename).__doc__,
                                        'missing description!')

            except KeyboardInterrupt:
                raise

            except:
                output('***** failed! *****')
                docstring = 'example import failed!'

            ifd.write('   %s <%s>\n' % (os.path.basename(ebase), rst_base))
            fig_include = ''
            for fig_filename in _get_fig_filenames(ebase, images_dir):
                rst_fig_filename = _make_sphinx_path(fig_filename)
                if os.path.exists(fig_filename):
                    fig_include += _image % rst_fig_filename + '\n'
                else:
                    output('   warning: figure "%s" not found' % fig_filename)

            # Example rst file.
            fd = open(full_rst_filename, 'w', encoding='utf-8')
            fd.write(_include % (rst_base, ebase, '=' * len(ebase),
                                 docstring,
                                 fig_include,
                                 rst_ex_filename, rst_ex_filename))
            fd.close()

        ifd.close()

        mfd.write('   %s-index\n' % dirname)

    mfd.close()

    output('...done')

    return dir_map


_rst_empty_item = """\
        - ..
"""

_rst_item = """\
      %s - .. figure:: %s
             :target: %s

             :ref:`%s <%s>`
"""

_gallery_table = """\
   .. list-table::
      :align: center
      :class: gallery
"""

_gallery_head = """\
.. only:: html

   .. _gallery-index:

   Gallery
   =======
"""


def generate_gallery(examples_dir, output_filename, doc_dir,
                     rst_dir, thumbnails_dir, dir_map, n_col=3):
    """
    Generate the gallery rst file with thumbnail images and links to
    examples.

    Parameters
    ----------
    output_filename : str
        The output rst file name.
    doc_dir : str
        The top level directory of gallery files.
    rst_dir : str
        The full path to rst files of examples within `doc_dir`.
    thumbnails_dir : str
        The full path to thumbnail images within `doc_dir`.
    dir_map : dict
        The directory mapping returned by `generate_rst_files()`
    n_col : int
        The number of columns in the gallery table.
    """
    output('generating %s...' % output_filename)

    lines = [_gallery_head]

    for dirname, filenames in ordered_iteritems(dir_map):
        title = ['   %s' % dirname.title().replace('_', ' '),
                 '   ' + len(dirname) * '^' + '',
                 _gallery_table]

        llines = []
        icol = 0
        for ex_filename, rst_filename in filenames:
            ebase = ex_filename.replace(examples_dir, '')[1:]
            link = os.path.splitext(rst_filename)[0]

            try:
                thumbnail_filename = next(_get_fig_filenames(ebase,
                                                             thumbnails_dir))
            except StopIteration:
                thumbnail_filename = ''

            if not os.path.isfile(thumbnail_filename):
                # Skip examples with no image (= failed examples).
                output('warning: figure "%s" not found' % thumbnail_filename)
                continue

            thumbnail_name = thumbnail_filename.replace(doc_dir, '..')
            path_to_file = os.path.join(examples_dir, ebase)
            docstring = get_default(import_file(path_to_file).__doc__,
                                    'missing description!')
            docstring = docstring.replace('e.g.', 'eg:')
            docstring = docstring.split('.')
            label = docstring[0].strip()
            label = label.replace('\n', ' ')
            label = label.replace('  ', ' ')

            llines.append(_rst_item % (' ' if icol else '*', thumbnail_name,
                                       link + '.html', label, link))
            icol = (icol + 1) % n_col

        if icol > 0:
            for j in range(icol, n_col):
                llines.append(_rst_empty_item)

        if len(llines) > 0:
            lines += title + llines
        else:
            output('warning: no figures in "%s"' % dirname)

    fd = open(output_filename, 'wt')
    fd.write('\n'.join(lines))
    fd.close()

    output('...done')


helps = {
    'doc_dir': 'top level directory of gallery files',
    'pattern': 'example files search pattern [default: %(default)s]',
    'no_images': 'do not (re)generate images',
    'no_thumbnails': 'do not (re)generate thumbnails',
    'output_filename': 'output file name [default: %(default)s]',
}


def main():
    parser = ArgumentParser(description=__doc__,
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('--version', action='version', version='%(prog)s')
    parser.add_argument('-d', '--doc-dir', metavar='doc_dir',
                        action='store', dest='doc_dir',
                        default='doc', help=helps['doc_dir'])
    parser.add_argument('-p', '--pattern', metavar='pattern',
                        action='store', dest='pattern',
                        default='*.py', help=helps['pattern'])
    parser.add_argument('-n', '--no-images',
                        action='store_false', dest='images',
                        default=True, help=helps['no_images'])
    parser.add_argument('--no-thumbnails',
                        action='store_false', dest='thumbnails',
                        default=True, help=helps['no_thumbnails'])
    parser.add_argument('-o', '--output', metavar='output_filename',
                        action='store', dest='output_filename',
                        default='gallery.rst',
                        help=helps['output_filename'])
    options = parser.parse_args()

    rst_dir = 'examples'
    doc_dir = os.path.realpath(options.doc_dir)
    examples_dir = os.path.realpath('sfepy/examples')
    full_rst_dir = os.path.join(doc_dir, rst_dir)
    images_dir = os.path.join(doc_dir, 'images/gallery')
    thumbnails_dir = os.path.join(images_dir, 'thumbnails')

    output_filename = os.path.join(full_rst_dir, options.output_filename)

    if options.images:
        generate_images(images_dir, examples_dir, pattern=options.pattern)

    if options.thumbnails:
        generate_thumbnails(thumbnails_dir, images_dir)

    dir_map = generate_rst_files(full_rst_dir, examples_dir, images_dir,
                                 pattern=options.pattern)

    generate_gallery(examples_dir, output_filename, doc_dir,
                     rst_dir, thumbnails_dir, dir_map)


if __name__ == '__main__':
    main()



from sfepy.base.base import assert_

def test_struct_add():
    from sfepy.base.base import Struct
    from copy import deepcopy

    a = Struct(f1=0,
               f2=[1, 2, 3],
               f3=Struct(ff='abc'),
               f4=3.14)
    a0 = deepcopy(a)
    b = Struct(f1=5,
               f2=[1],
               f3=Struct(ff='', gg=123),
               f5='new one')
    c = a + b

    assert_(c.f1 == 0)
    assert_(c.f2 == [1, 2, 3])
    assert_(c.f3.ff == 'abc')
    assert_(c.f3.gg == 123)
    assert_(c.f4 == 3.14)
    assert_(c.f5 == 'new one')

    assert_(a.f1 == a0.f1)
    assert_(a.f2 == a0.f2)
    assert_(a.f3.ff == a0.f3.ff)
    assert_(a.f4 == a0.f4)

def test_struct_i_add():
    from sfepy.base.base import Struct

    a = Struct(f1=0,
               f2=[1, 2, 3],
               f3=Struct(ff='abc'))
    b = Struct(f1=5,
               f2=[1],
               f3=Struct(ff='', gg=123),
               f4='new one')
    a += b

    assert_(a.f1 == 0)
    assert_(a.f2 == [1, 2, 3])
    assert_(a.f3.ff == 'abc')
    assert_(a.f3.gg == 123)
    assert_(a.f4 == 'new one')

def test_container_add():
    from sfepy.base.base import Struct, Container

    a = Struct(name='a')
    b = Struct(name='b')

    c1 = Container()

    c1 = c1 + c1
    assert_(c1.names == [])

    c1 += Container([a, b])
    assert_(c1.names == ['a', 'b'])

    c2 = c1 + c1
    assert_(c2.names == 2 * ['a', 'b'])

    c2 += c2
    assert_(c2.names == 4 * ['a', 'b'])

def test_verbose_output():
    from io import StringIO

    from sfepy.base.base import Output, goptions

    fd = StringIO()

    output = Output('test', filename=fd)

    output('test1')
    goptions['verbose'] = False
    output('test2')
    goptions['verbose'] = 1
    output('test3')

    _ok1 = bool(goptions['verbose'])
    _ok2 = fd.getvalue() == 'test test1\ntest test3\n'

    fd.close()

    ok = _ok1 and _ok2

    assert_(ok)

def test_resolve_deps():
    from sfepy.base.resolve_deps import resolve

    deps = {
        'a' : ['a', 'b'],
        'b' : ['a', 'b'],
        'c' : ['b', 'c', 'd', 'e'],
        'd' : ['c', 'e'],
        'e' : ['d', 'e'],
        'f' : ['e', 'f', 'g'],
        'g' : ['g'],
    }
    order = resolve(deps)
    ok1 = order == [['g'], ['a', 'b'], ['c', 'd', 'e'], ['f']]

    deps = {
        'a' : ['b'],
        'b' : ['c'],
        'c' : ['a'],
    }
    order = resolve(deps)
    ok2 = order == [['a', 'b', 'c']]

    assert_(ok1 and ok2)

def test_parse_conf():
    from sfepy.base.parse_conf import list_dict
    from sfepy.base.conf import dict_from_string as parse_dict
    ld = list_dict()

    def parse(x):
        out = ld.parseString(x, True)[0]
        return out

    assert_(parse("1,2") == ([1,2],{}))
    assert_(parse("1,[2]") == ([1,[2]],{}))
    assert_(parse("1,[2,4],c=3") == ([1,[2,4]],{'c':3}))
    assert_(parse("1,(2),c:3,uu=7") == ([1,(2,)],{'c':3,'uu':7}))
    assert_(parse("'long string ([\"',(2,5),c:3") ==
                 (['long string (["',(2,5)],{'c':3}))


    assert_(parse_dict('') == {})
    assert_(parse_dict('a:[]') == {'a':[]})
    assert_(parse_dict('a:{}') == {'a':{}})
    assert_(parse_dict('1:2,a:{},3:4') == {1:2,'a':{},3:4})

    assert_(parse('') == ([],{}))
    assert_(parse('[ ]') == ([[]],{}))
    assert_(parse('[]') == ([[]],{}))
    assert_(parse('[[]]') == ([[[]]],{}))
    assert_(parse('[[[]]]')==([[[[]]]],{}))
    assert_(parse('a,{},[],None,True,False,"False"') ==
                     (['a',{},[],None,True,False,"False"],{}))



from copy import copy

import numpy as nm
import pytest

from sfepy.base.base import ordered_iteritems
from sfepy import data_dir
import sfepy.base.testing as tst

filename_meshes = [data_dir + '/meshes/elements/%s_2.mesh' % geom
                   for geom in ['1_2', '2_3', '2_4', '3_4', '3_8', '3_2_4']]

not_tested_terms = ['dw_ns_dot_grad_s',
                    'dw_lin_spring', 'dw_lin_truss', 'ev_lin_truss_force',
                    'dw_tl_he_ogden_ad']


def make_term_args(arg_shapes, arg_kinds, arg_types, ats_mode, domain,
                   material_value=None, poly_space_base=None):
    from sfepy.base.base import basestr
    from sfepy.discrete import FieldVariable, Material, Variables, Materials
    from sfepy.discrete.fem import Field
    from sfepy.solvers.ts import TimeStepper
    from sfepy.mechanics.tensors import dim2sym

    omega = domain.regions['Omega']
    dim = domain.shape.dim
    sym = dim2sym(dim)

    def _parse_scalar_shape(sh):
        if isinstance(sh, basestr):
            if sh == 'D':
                return dim

            elif sh == 'D2':
                return dim**2

            elif sh == 'S':
                return sym

            elif sh == 'N': # General number ;)
                return 1

            elif sh == 'str':
                return 'str'

            else:
                return int(sh)

        else:
            return sh

    def _parse_tuple_shape(sh):
        if isinstance(sh, basestr):
            return [_parse_scalar_shape(ii.strip()) for ii in sh.split(',')]

        else:
            return (int(sh),)

    args = {}
    str_args = []
    materials = []
    variables = []
    for ii, arg_kind in enumerate(arg_kinds):
        if arg_kind != 'ts':
            if ats_mode is not None:
                extended_ats = arg_types[ii] + ('/%s' % ats_mode)

            else:
                extended_ats = arg_types[ii]

            try:
                sh = arg_shapes[arg_types[ii]]

            except KeyError:
                sh = arg_shapes[extended_ats]

        if arg_kind.endswith('variable'):
            shape = _parse_scalar_shape(sh[0] if isinstance(sh, tuple) else sh)
            field = Field.from_args('f%d' % ii, nm.float64, shape, omega,
                                    approx_order=1,
                                    poly_space_base=poly_space_base)

            if arg_kind == 'virtual_variable':
                if sh[1] is not None:
                    istate = arg_types.index(sh[1])

                else:
                    # Only virtual variable in arguments.
                    istate = -1
                    # -> Make fake variable.
                    var = FieldVariable('u-1', 'unknown', field)
                    var.set_constant(0.0)
                    variables.append(var)

                var = FieldVariable('v', 'test', field,
                                    primary_var_name='u%d' % istate)

            elif arg_kind == 'state_variable':
                var = FieldVariable('u%d' % ii, 'unknown', field)
                var.set_constant(0.0)

            elif arg_kind == 'parameter_variable':
                var = FieldVariable('p%d' % ii, 'parameter', field,
                                    primary_var_name='(set-to-None)')
                var.set_constant(0.0)

            variables.append(var)
            str_args.append(var.name)
            args[var.name] = var

        elif arg_kind.endswith('material'):
            if sh is None: # Switched-off opt_material.
                continue

            prefix = ''
            if isinstance(sh, basestr):
                aux = sh.split(':')
                if len(aux) == 2:
                    prefix, sh = aux

            value = material_value
            if value is None:
                value = 1.0

            elif isinstance(value, dict):
                value = material_value.get(arg_types[ii], 1.0)

            shape = _parse_tuple_shape(sh)
            if 'str' in shape:
                values = {'.c%d' % ii : value}

            elif (len(shape) > 1) or (shape[0] > 1):
                if ((len(shape) == 2) and (shape[0] ==  shape[1])
                    and (value != 0.0)):
                    # Identity matrix.
                    val = nm.eye(shape[0], dtype=nm.float64)

                else:
                    # Array.
                    val = nm.empty(shape, dtype=nm.float64)
                    val.fill(value)

                values = {'%sc%d' % (prefix, ii)
                          : val}

            elif (len(shape) == 1) and (shape[0] == 1):
                # Single scalar as a special value.
                values = {'.c%d' % ii : value}

            else:
                raise ValueError('wrong material shape! (%s)' % shape)

            mat = Material('m%d' % ii, values=values)

            materials.append(mat)
            str_args.append(mat.name + '.' + 'c%d' % ii)
            args[mat.name] = mat

        elif arg_kind == 'ts':
            ts = TimeStepper(0.0, 1.0, 1.0, 5)
            str_args.append('ts')
            args['ts'] = ts

        else:
            str_args.append('user%d' % ii)
            args[str_args[-1]] = sh if callable(sh) else None

    materials = Materials(materials)
    variables = Variables(variables)

    return args, str_args, materials, variables

@pytest.fixture(scope='module')
def data():
    from sfepy.base.base import Struct
    from sfepy.discrete import Integral
    from sfepy.discrete.fem import Mesh, FEDomain

    domains = []
    for filename in filename_meshes:
        mesh = Mesh.from_file(filename)
        domain = FEDomain('domain_%s' % mesh.name.replace(data_dir, ''),
                          mesh)
        domain.create_region('Omega', 'all')
        domain.create_region('Gamma', 'vertices of surface', 'facet')

        domains.append(domain)

    integral = Integral('i', order=3)
    qp_coors, qp_weights = integral.get_qp('3_8')
    custom_integral = Integral('i', coors=qp_coors, weights=qp_weights,
                               order='custom')

    return Struct(domains=domains, integral=integral,
                  custom_integral=custom_integral)

def _test_single_term(data, term_cls, domain, rname):
    from sfepy.terms import Term
    from sfepy.terms.terms import get_arg_kinds

    ok = True

    term_call = term_cls.name + '(%s)'

    arg_shapes_list = term_cls.arg_shapes
    if not isinstance(arg_shapes_list, list):
        arg_shapes_list = [arg_shapes_list]

    if term_cls.integration != 'custom':
        integral = data.integral

    else:
        integral = data.custom_integral

    poly_space_base = getattr(term_cls, 'poly_space_base', 'lagrange')

    prev_shapes = {}
    for _arg_shapes in arg_shapes_list:
        # Unset shapes are taken from the previous iteration.
        arg_shapes = copy(prev_shapes)
        arg_shapes.update(_arg_shapes)
        prev_shapes = arg_shapes

        tst.report('arg_shapes:', arg_shapes)
        arg_types = term_cls.arg_types
        if not isinstance(arg_types[0], tuple):
            arg_types = (arg_types,)

        for iat, ats in enumerate(arg_types):
            tst.report('arg_types:', ats)

            arg_kinds = get_arg_kinds(ats)
            modes = getattr(term_cls, 'modes', None)
            mode = modes[iat] if modes is not None else None

            if 'dw_s_dot_grad_i_s' in term_cls.name:
                material_value = 0.0

            elif 'de_mass' in term_cls.name:
                material_value = {'material_lumping' : 'row_sum'}

            else:
                material_value = 1.0
            aux = make_term_args(arg_shapes, arg_kinds, ats, mode, domain,
                                 material_value=material_value,
                                 poly_space_base=poly_space_base)
            args, str_args, materials, variables = aux

            tst.report('args:', str_args)

            name = term_call % (', '.join(str_args))
            term = Term.new(name, integral, domain.regions[rname], **args)
            term.setup()

            call_mode = 'weak' if term.names.virtual else 'eval'
            tst.report('call mode:', call_mode)

            out = term.evaluate(mode=call_mode, ret_status=True)

            if call_mode == 'eval':
                vals, status = out
                vals = nm.array(vals)

            else:
                vals, iels, status = out

            if isinstance(vals, tuple):
                # Dynamic connectivity terms.
                vals = vals[0]

            _ok = nm.isfinite(vals).all()
            ok = ok and _ok
            tst.report('values shape: %s' % (vals.shape,))
            if not _ok:
                tst.report('values are not finite!')
                tst.report(vals)

            _ok = status == 0
            if not _ok:
                tst.report('status is %s!' % status)

            ok = ok and _ok

            if term.names.virtual:
                # Test differentiation w.r.t. state variables in the weak
                # mode.
                svars = term.get_state_variables(unknown_only=True)
                for svar in svars:
                    vals, iels, status = term.evaluate(mode=call_mode,
                                                       diff_var=svar.name,
                                                       ret_status=True)
                    if isinstance(vals, tuple):
                        # Dynamic connectivity terms.
                        vals = vals[0]

                    _ok = status == 0
                    ok = ok and _ok
                    tst.report('diff: %s' % svar.name)
                    if not _ok:
                        tst.report('status is %s!' % status)

                    _ok = nm.isfinite(vals).all()
                    ok = ok and _ok
                    tst.report('values shape: %s' % (vals.shape,))
                    if not _ok:
                        tst.report('values are not finite!')
                        tst.report(vals)

    return ok

@pytest.mark.slow
def test_term_call_modes(data):
    from sfepy.terms import term_table
    ok = True

    failed = []
    for domain in data.domains:
        tst.report('domain: %s' % domain.name)

        domain_geometry = list(domain.geom_els.values())[0].name
        if domain.shape.dim != domain.shape.tdim:
            domain_geometry = '%d_%s' % (domain.shape.dim, domain_geometry)

        for _, term_cls in ordered_iteritems(term_table):
            if (domain_geometry not in term_cls.geometries) \
               or ("dg" in term_cls.name) \
               or (term_cls.name in not_tested_terms):
                continue

            rnames = []
            if ('cell' in term_cls.integration or
                'point' in term_cls.integration or
                'custom' in term_cls.integration):
                rnames.append('Omega')

            if ('facet' in term_cls.integration or
                'facet_extra' in term_cls.integration):
                rnames.append('Gamma')

            for rname in rnames:
                tst.report('<-- %s.%s ...' % (term_cls.name, rname))

                if rname == 'Gamma' and domain.mesh.dim == 1:
                    tst.report('--> 1D Gamma region: not tested!')

                elif term_cls.arg_shapes:
                    try:
                        _ok = _test_single_term(data, term_cls, domain,
                                                rname)

                    except:
                        _ok = False

                    if not _ok:
                        failed.append((domain.name, term_cls.name))

                    ok = ok and _ok
                    tst.report('--> ok: %s' % _ok)

                else:
                    tst.report('--> not tested!')

    tst.report('failed:', failed)

    assert ok



"""
Live plot demonstration.

Usage Example
-------------

- Run and plot two logs on the fly::

    python3 sfepy/examples/miscellaneous/live_plot.py --plot-log

- Run and store the two logs, plot them later::

    python3 sfepy/examples/miscellaneous/live_plot.py

    python3 sfepy/scripts/plot_logs.py live_plot.txt
    python3 sfepy/scripts/plot_logs.py live_plot2.txt
"""
from argparse import ArgumentParser, RawDescriptionHelpFormatter
import os.path as op
from functools import partial

import numpy as nm

from sfepy.base.base import output, pause
from sfepy.base.log import Log

def main():
    parser = ArgumentParser(description=__doc__.rstrip(),
                            formatter_class=RawDescriptionHelpFormatter)
    parser.add_argument('-o', '--output-dir', default='.',
                        help='output directory')
    parser.add_argument( '-p', '--plot-log', action='store_true',
                         default=False, help='show live log figures')
    options = parser.parse_args()

    inodir = partial(op.join, options.output_dir)

    log = Log((['sin(x) + i sin(x**2)', 'cos(x)'], ['exp(x)'], ['x^2']),
              yscales=['linear', 'log', 'linear'],
              xlabels=['angle', None, 'x'],
              ylabels=[None, 'a function', 'square'],
              aggregate=1000, sleep=0.05,
              is_plot=options.plot_log,
              log_filename=inodir('live_plot.txt'))

    # Plot groups can be added on the fly but then they are missing in the log
    # header - sfepy/scripts/plot_logs.py cannot plot them.
    # log.add_group(['x^2'], yscale='linear', xlabel='new x',
    #               ylabel='square', formats=['%+g'])

    log2 = Log([['x^3']],
               yscales=['linear'],
               xlabels=['x'], ylabels=['a cubic function'],
               aggregate=1000, sleep=0.05,
               is_plot=options.plot_log,
               log_filename=inodir('live_plot2.txt'),
               formats=[['{:.5e}']])

    for ii, x in enumerate(nm.linspace(0, 4.0 * nm.pi, 200)):
        output('x: ', x)

        log(nm.sin(x)+1j*nm.sin(x**2), nm.cos(x), nm.exp(x), x**2,
            x=[x, None, x])

        if (ii == 20) or (ii == 50):
            log.plot_vlines([2], color='g', linewidth=2)

        log2(x**3, x=[x])

    print(log)
    print(log2)

    log(save_figure=inodir('live_plot.png'))
    log2(save_figure=inodir('live_plot2.png'))

    if options.plot_log:
        pause()

    log(finished=True)
    log2(finished=True)

if __name__ == '__main__':
    main()



# coding=utf8
import os
import os.path as op

from sfepy import data_dir
from sfepy.base.base import assert_
import sfepy.base.testing as tst

filename_meshes = ['/meshes/3d/cylinder.mesh',
                   '/meshes/3d/cylinder.vtk',
                   '/meshes/various_formats/small2d.mesh',
                   '/meshes/various_formats/small2d.vtk',
                   '/meshes/various_formats/octahedron.node',
                   '/meshes/various_formats/comsol_tri.txt',
                   '/meshes/various_formats/abaqus_hex.inp',
                   '/meshes/various_formats/abaqus_tet.inp',
                   '/meshes/various_formats/abaqus_quad.inp',
                   '/meshes/various_formats/abaqus_tri.inp',
                   '/meshes/various_formats/abaqus_quad_tri.inp',
                   '/meshes/various_formats/hex4.mesh3d',
                   '/meshes/various_formats/tetra8.mesh3d',
                   '/meshes/various_formats/cube.bdf',
                   '/meshes/various_formats/msh_tri.msh',
                   '/meshes/various_formats/msh_tetra.msh',
                   '/meshes/various_formats/xyz_quad.xyz',
                   '/meshes/various_formats/xyz_tet.xyz']
filename_meshes = [data_dir + name for name in filename_meshes]

def mesh_hook(mesh, mode):
    """
    Define a mesh programmatically.
    """
    if mode == 'read':
        nodes = [[0, 0], [1, 0], [1, 1], [0, 1]]
        nod_ids = [0, 0, 1, 1]
        conns = [[[0, 1, 2], [0, 2, 3]]]
        mat_ids = [[0, 1]]
        descs = ['2_3']

        mesh._set_io_data(nodes, nod_ids, conns, mat_ids, descs)

    elif mode == 'write':
        pass

from sfepy.discrete.fem.meshio import UserMeshIO

filename_meshes.extend([mesh_hook, UserMeshIO(mesh_hook)])

same = [(0, 1), (2, 3)]

def _compare_meshes(mesh0, mesh1, flag='cv'):
    import numpy as nm

    oks = []

    ok0 = (mesh0.dim == mesh1.dim)
    if not ok0:
        tst.report('dimension failed!')
    oks.append(ok0)

    ok0 = mesh0.n_nod == mesh1.n_nod
    if not ok0:
        tst.report('number of nodes failed!')
    oks.append(ok0)

    ok0 = mesh0.n_el == mesh1.n_el
    if not ok0:
        tst.report('number of elements failed!')
    oks.append(ok0)

    ok0 = mesh0.descs == mesh1.descs
    if not ok0:
        tst.report('element types failed!')
    oks.append(ok0)

    ok0 = nm.allclose(mesh0.coors, mesh1.coors)
    if not ok0:
        tst.report('nodes failed!')
    oks.append(ok0)

    if 'v' in flag:
        ok0 = nm.all(mesh0.cmesh.vertex_groups == mesh1.cmesh.vertex_groups)
        if not ok0:
            tst.report('node groups failed!')
        oks.append(ok0)

    if 'c' in flag:
        ok0 = nm.all(mesh0.cmesh.cell_groups == mesh1.cmesh.cell_groups)
        if not ok0:
            tst.report('material ids failed!')
        oks.append(ok0)

    ok0 = (nm.all(mesh0.cmesh.get_cell_conn().indices ==
                  mesh1.cmesh.get_cell_conn().indices) and
           nm.all(mesh0.cmesh.get_cell_conn().offsets ==
                  mesh1.cmesh.get_cell_conn().offsets))
    if not ok0:
        tst.report('connectivities failed!')
    oks.append(ok0)

    return oks

def test_read_meshes():
    """Try to read all listed meshes."""
    from sfepy.discrete.fem import Mesh

    conf_dir = op.dirname(__file__)
    oks = []
    for ii, filename in enumerate(filename_meshes):
        tst.report('%d. mesh: %s' % (ii + 1, filename))
        try:
            mesh = Mesh.from_file(filename, prefix_dir=conf_dir)

        except Exception as exc:
            tst.report(exc)
            tst.report('read failed!')
            oks.append(False)
            continue

        try:
            assert_(mesh.dim == (mesh.coors.shape[1]))
            assert_(mesh.n_nod == (mesh.coors.shape[0]))
            assert_(mesh.n_nod == (mesh.cmesh.vertex_groups.shape[0]))
            assert_(mesh.n_el == mesh.cmesh.num[mesh.cmesh.tdim])

        except ValueError as exc:
            tst.report(exc)
            tst.report('read assertion failed!')
            oks.append(False)
            continue

        oks.append(True)
        tst.report('read ok')

    assert_(all(oks))

def test_compare_same_meshes():
    """
    Compare same meshes in various formats.
    """
    from sfepy.discrete.fem import Mesh

    conf_dir = op.dirname(__file__)
    oks = []
    for i0, i1 in same:
        name0 = filename_meshes[i0]
        name1 = filename_meshes[i1]
        tst.report('comparing meshes from "%s" and "%s"' % (name0, name1))

        mesh0 = Mesh.from_file(name0, prefix_dir=conf_dir)
        mesh1 = Mesh.from_file(name1, prefix_dir=conf_dir)
        ok = _compare_meshes(mesh0, mesh1)
        tst.report('->', ok)
        oks.extend(ok)

    assert_(all(oks))

def test_read_dimension():
    from sfepy.discrete.fem import MeshIO

    meshes = {data_dir + '/meshes/various_formats/small2d.mesh' : 2,
              data_dir + '/meshes/various_formats/small2d.vtk' : 2,
              data_dir + '/meshes/various_formats/small3d.mesh' : 3}

    ok = True
    conf_dir = op.dirname(__file__)
    for filename, adim in meshes.items():
        tst.report('mesh: %s, dimension %d' % (filename, adim))
        io = MeshIO.any_from_filename(filename, prefix_dir=conf_dir)
        dim = io.read_dimension()
        if dim != adim:
            tst.report('read dimension %d -> failed' % dim)
            ok = False
        else:
            tst.report('read dimension %d -> ok' % dim)

    assert_(ok)

def test_write_read_meshes(output_dir):
    """
    Try to write and then read all supported formats.
    """
    import numpy as nm
    from sfepy.discrete.fem import Mesh
    from sfepy.discrete.fem.meshio import supported_formats

    conf_dir = op.dirname(__file__)
    mesh0 = Mesh.from_file(data_dir
                           + '/meshes/various_formats/small3d.mesh',
                           prefix_dir=conf_dir)

    mesh0.cmesh.vertex_groups[:] =\
        nm.random.randint(1, 10, size=mesh0.cmesh.n_coor)

    mesh0.cmesh.cell_groups[:] =\
        nm.random.randint(1, 10, size=mesh0.cmesh.n_el)

    oks = []
    for name, (cls, suffix, flag) in supported_formats.items():
        if 'w' not in flag: continue

        suffix = suffix[0]  # only the first of possible suffixes
        filename = op.join(output_dir, 'test_mesh_wr' + suffix)
        tst.report('%s format: %s' % (suffix, filename))

        try:
            mesh0.write(filename, io='auto')

        except:
            if cls == 'meshio':
                import traceback
                tst.report('-> cannot write "%s" format into "%s",'
                           ' skipping' % (name, filename))
                tb = traceback.format_exc()
                tst.report('reason:\n', tb)
                continue

            else:
                raise

        else:
            mesh1 = Mesh.from_file(filename)

        ok = _compare_meshes(mesh0, mesh1, flag)
        tst.report('->', ok)
        oks.extend(ok)

    assert_(all(oks))

def test_hdf5_meshio():
    try:
        from igakit import igalib; igalib
    except ImportError:
        tst.report('hdf5_meshio not-tested (missing igalib module)!')
        return

    import tempfile
    import numpy as nm
    import scipy.sparse as sps
    from sfepy.discrete.fem.meshio import HDF5MeshIO
    from sfepy.base.base import Struct
    from sfepy.base.ioutils import Cached, Uncached, SoftLink, \
                                   DataSoftLink
    from sfepy.discrete.iga.domain import IGDomain
    from sfepy.discrete.iga.domain_generators import gen_patch_block_domain
    from sfepy.solvers.ts import TimeStepper
    from sfepy.discrete.fem import Mesh

    conf_dir = op.dirname(__file__)
    mesh0 = Mesh.from_file(data_dir +
                           '/meshes/various_formats/small3d.mesh',
                           prefix_dir=conf_dir)

    shape = [4, 4, 4]
    dims = [5, 5, 5]
    centre = [0, 0, 0]
    degrees = [2, 2, 2]

    nurbs, bmesh, regions = gen_patch_block_domain(dims, shape, centre,
                                                   degrees,
                                                   cp_mode='greville',
                                                   name='iga')
    ig_domain = IGDomain('iga', nurbs, bmesh, regions=regions)

    int_ar = nm.arange(4)

    data = {
        'list': range(4),
        'mesh1': mesh0,
        'mesh2': mesh0,
        'mesh3': Uncached(mesh0),
        'mesh4': SoftLink('/step0/__cdata/data/data/mesh2'),
        'mesh5': DataSoftLink('Mesh','/step0/__cdata/data/data/mesh1/data'),
        'mesh6': DataSoftLink('Mesh','/step0/__cdata/data/data/mesh2/data',
            mesh0),
        'mesh7': DataSoftLink('Mesh','/step0/__cdata/data/data/mesh1/data',
            True),
        'iga' : ig_domain,
        'cached1': Cached(1),
        'cached2': Cached(int_ar),
        'cached3': Cached(int_ar),
        'types': ( True, False, None ),
        'tuple': ('first string', 'druhý UTF8 řetězec'),
        'struct': Struct(
            double=nm.arange(4, dtype=float),
            int=nm.array([2,3,4,7]),
            sparse=sps.csr_matrix(nm.array([1,0,0,5]).
                                  reshape((2,2)))
         )
    }

    with tempfile.NamedTemporaryFile(suffix='.h5', delete=False) as fil:
        io = HDF5MeshIO(fil.name)
        ts = TimeStepper(0,1.,0.1, 10)

        io.write(fil.name, mesh0, {
            'cdata' : Struct(
                mode='custom',
                data=data,
                unpack_markers=False
            )
        }, ts=ts)
        ts.advance()

        mesh = io.read()
        data['problem_mesh'] = DataSoftLink('Mesh', '/mesh', mesh)

        io.write(fil.name, mesh0, {
            'cdata' : Struct(
                mode='custom',
                data=data,
                unpack_markers=True
            )
        }, ts=ts)

        cache = {'/mesh': mesh }
        fout = io.read_data(0, cache=cache)
        fout2 = io.read_data(1, cache=cache )
        out = fout['cdata']
        out2 = fout2['cdata']

        assert_(out['mesh7'] is out2['mesh7'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh6'] is out2['mesh6'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh5'] is not out2['mesh5'],
                'These two meshes shouldn''t be in fact the same object')

        assert_(out['mesh1'] is out['mesh2'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh1'] is out['mesh2'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh4'] is out['mesh2'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh5'] is not out['mesh2'],
                'These two meshes shouldn''t be in fact the same object')

        assert_(out['mesh6'] is out['mesh2'],
                'These two meshes should be in fact the same object')

        assert_(out['mesh7'] is not out['mesh2'],
                'These two meshes shouldn''t be in fact the same object')

        assert_(out['mesh3'] is not out['mesh2'],
                'These two meshes should be different objects')

        assert_(out['cached2'] is out['cached3'],
                'These two array should be the same object')

        assert_(out2['problem_mesh'] is mesh,
                'These two meshes should be the same objects')

        assert_(_compare_meshes(out['mesh1'], mesh0),
                'Failed to restore mesh')

        assert_(_compare_meshes(out['mesh3'], mesh0),
                'Failed to restore mesh')

        assert_((out['struct'].sparse.todense()
                 == data['struct'].sparse.todense()).all(),
                'Sparse matrix restore failed')

        ts.advance()
        io.write(fil.name, mesh0, {
                'cdata' : Struct(
                    mode='custom',
                    data=[
                        DataSoftLink('Mesh',
                                     '/step0/__cdata/data/data/mesh1/data',
                                     mesh0),
                        mesh0
                    ]
                )
        }, ts=ts)
        out3 = io.read_data(2)['cdata']
        assert_(out3[0] is out3[1])

    os.remove(fil.name)

    #this property is not restored
    del data['iga'].nurbs.nurbs

    #not supporting comparison
    del data['iga']._bnf
    del out2['iga']._bnf

    #restoration of this property fails
    del data['iga'].vertex_set_bcs
    del out2['iga'].vertex_set_bcs

    #these soflinks have no information how to unpack, so it must be
    #done manually
    data['mesh4'] = mesh0
    data['mesh5'] = mesh0
    data['mesh7'] = mesh0

    for key, val in out2.items():
        tst.report('comparing:', key)
        tst.assert_equal(val, data[key])



import sfepy.base.testing as tst

def test_elastic_constants():
    import numpy as nm
    from sfepy.mechanics.matcoefs import ElasticConstants

    ok = True

    names = ['bulk', 'lam', 'mu', 'young', 'poisson', 'p_wave']

    ec = ElasticConstants(lam=1.0, mu=1.5)
    vals = ec.get(names)

    tst.report('using values:', vals)

    for i1 in range(len(names)):
        for i2 in range(i1+1, len(names)):
            kwargs = {names[i1] : vals[i1], names[i2] : vals[i2]}

            try:
                ec.init(**kwargs)

            except:
                _ok = False

            else:
                _ok = True

            ec_vals = ec.get(names)
            _ok = _ok and nm.allclose(ec_vals, vals)

            tst.report(names[i1], names[i2], '->', _ok)
            if not _ok:
                tst.report('correct:', vals)
                tst.report('    got:', ec_vals)

            ok = ok and _ok

    assert ok

def test_conversion_functions():
    import numpy as nm

    import sfepy.mechanics.matcoefs as mc

    ok = True

    lam = 1.0
    mu = 1.5

    ec = mc.ElasticConstants(lam=lam, mu=mu)
    young, poisson, bulk = ec.get(['young', 'poisson', 'bulk'])

    lam = nm.array([lam] * 3)
    mu = nm.array([mu] * 3)
    young = nm.array([young] * 3)
    poisson = nm.array([poisson] * 3)

    _lam, _mu = mc.lame_from_youngpoisson(young, poisson)
    _ok = (nm.allclose(lam, _lam, rtol=0.0, atol=1e-14) and
           nm.allclose(mu, _mu, rtol=0.0, atol=1e-14))
    tst.report('lame_from_youngpoisson():', _ok)
    if not _ok:
        tst.report('correct:', lam, mu)
        tst.report('    got:', _lam, _mu)
    ok = ok and _ok

    _bulk = mc.bulk_from_youngpoisson(young, poisson)
    _ok = nm.allclose(bulk, _bulk, rtol=0.0, atol=1e-14)
    tst.report('bulk_from_youngpoisson():', _ok)
    if not _ok:
        tst.report('correct:', bulk)
        tst.report('    got:', _bulk)
    ok = ok and _ok

    _bulk = mc.bulk_from_lame(lam, mu)
    _ok = nm.allclose(bulk, _bulk, rtol=0.0, atol=1e-14)
    tst.report('bulk_from_lame():', _ok)
    if not _ok:
        tst.report('correct:', bulk)
        tst.report('    got:', _bulk)
    ok = ok and _ok

    assert ok

def test_stiffness_tensors():
    import numpy as nm

    from sfepy.base.base import assert_
    import sfepy.mechanics.matcoefs as mc

    ok = True

    lam = 1.0
    mu = 4.0

    lam = nm.array([lam] * 3)
    mu = nm.array([mu] * 3)

    d = nm.array([[ 9.,  1.,  1.,  0.,  0.,  0.],
                  [ 1.,  9.,  1.,  0.,  0.,  0.],
                  [ 1.,  1.,  9.,  0.,  0.,  0.],
                  [ 0.,  0.,  0.,  4.,  0.,  0.],
                  [ 0.,  0.,  0.,  0.,  4.,  0.],
                  [ 0.,  0.,  0.,  0.,  0.,  4.]])

    _ds = mc.stiffness_from_lame(3, lam, mu)
    assert_(_ds.shape == (3, 6, 6))

    _ok = True
    for _d in _ds:
        __ok = nm.allclose(_d, d, rtol=0.0, atol=1e-14)
        _ok = _ok and __ok
    tst.report('stiffness_from_lame():', _ok)
    ok = ok and _ok

    _lam, _mu = mc.lame_from_stiffness(d)
    _ok = (_lam == 1) and (_mu == 4)
    tst.report('lame_from_stiffness():', _ok)
    ok = ok and _ok

    young = 1.0
    poisson = 0.25

    d = mc.stiffness_from_youngpoisson(3, young, poisson, plane='strain')
    _young, _poisson = mc.youngpoisson_from_stiffness(d, plane='strain')
    _ok = nm.allclose([young, poisson], [_young, _poisson],
                      rtol=0.0, atol=1e-14)
    tst.report('youngpoisson_from_stiffness(plane="strain"):', _ok)
    ok = ok and _ok

    d = mc.stiffness_from_youngpoisson(2, young, poisson, plane='stress')
    _young, _poisson = mc.youngpoisson_from_stiffness(d, plane='stress')
    _ok = nm.allclose([young, poisson], [_young, _poisson],
                      rtol=0.0, atol=1e-14)
    tst.report('youngpoisson_from_stiffness(plane="stress"):', _ok)
    ok = ok and _ok

    d = 4.0 / 3.0 * nm.array([[ 4., -2., -2.,  0.,  0.,  0.],
                              [-2.,  4., -2.,  0.,  0.,  0.],
                              [-2., -2.,  4.,  0.,  0.,  0.],
                              [ 0.,  0.,  0.,  3.,  0.,  0.],
                              [ 0.,  0.,  0.,  0.,  3.,  0.],
                              [ 0.,  0.,  0.,  0.,  0.,  3.]])

    _ds = mc.stiffness_from_lame_mixed(3, lam, mu)
    assert_(_ds.shape == (3, 6, 6))

    _ok = True
    for _d in _ds:
        __ok = nm.allclose(_d, d, rtol=0.0, atol=1e-14)
        _ok = _ok and __ok
    tst.report('stiffness_from_lame_mixed():', _ok)
    ok = ok and _ok

    blam = - mu * 2.0 / 3.0
    _ds = mc.stiffness_from_lame(3, blam, mu)
    assert_(_ds.shape == (3, 6, 6))

    _ok = True
    for _d in _ds:
        __ok = nm.allclose(_d, d, rtol=0.0, atol=1e-14)
        _ok = _ok and __ok
    tst.report('stiffness_from_lame() with modified lambda:', _ok)
    ok = ok and _ok

    assert ok

def test_wave_speeds():
    import numpy as nm
    import sfepy.mechanics.matcoefs as mc

    E = 1e10
    nu = 0.3
    rho = 1100

    vp, vs = mc.wave_speeds_from_youngpoisson(E, nu, rho)
    E2, nu2 = mc.youngpoisson_from_wave_speeds(vp, vs, rho)

    tst.report(E, nu, ' == ', E2, nu2)

    assert nm.allclose(E, E2, rtol=1e-14, atol=0.0)
    assert nm.allclose(nu, nu2, rtol=1e-14, atol=0.0)



r"""
Nonlinear Poisson's equation example demonstrating the nonlinear diffusion
and nonlinear volume force terms.

The example is an adaptation of:
:ref:`diffusion-poisson_field_dependent_material`.

Find :math:`T` such that:

.. math::
   \int_{\Omega} c(T) \nabla s \cdot \nabla T + \int_{\Omega} g(T) \cdot s
    = 0
    \;, \quad \forall s \;.

where :math:`c(T)` and :math:`g(T)`  are the :math:`T` dependent coefficients.
"""
from sfepy import data_dir
import numpy as np

filename_mesh = data_dir + '/meshes/3d/cylinder.mesh'

def get_conductivity(T):
    """
    Calculates the conductivity as (1 + T**4) and returns it.
    """
    val = 1 + T**4
    return val

def d_get_conductivity(T):
    """
    Calculates the derivative of get_conductivity and returns it.
    """
    val = 4*T**3
    return val

def nl_src(T):
    """
    Calculates a non linear source term as np.sinh(T) and returns it.
    """
    val = np.sinh(T)
    return val

def d_nl_src(T):
    """
    Calculates the derivative of the nl_src term and returns it.
    """
    val = np.cosh(T)
    return val

materials = {
}

fields = {
    'temperature' : ('real', 1, 'Omega', 1),
}

variables = {
    'T' : ('unknown field', 'temperature', 0),
    's' : ('test field',    'temperature', 'T'),
}

regions = {
    'Omega' : 'all',
    'Gamma_Left' : ('vertices in (x < 0.00001)', 'facet'),
    'Gamma_Right' : ('vertices in (x > 0.099999)', 'facet'),
}

ebcs = {
    'T1' : ('Gamma_Left', {'T.0' : 2.0}),
    'T2' : ('Gamma_Right', {'T.0' : -2.0}),
}

integrals = {
    'i' : 10,
}

equations = {
    'Temperature' : """
       dw_nl_diffusion.i.Omega(get_conductivity, d_get_conductivity, s, T)
     + dw_volume_nvf.i.Omega(nl_src, d_nl_src, s, T) = 0
     """
}

solvers = {
    'ls' : ('ls.scipy_direct', {}),
    'newton' : ('nls.newton', {
        'i_max' : 15,
        'eps_a' : 1e-10,
        'eps_r' : 1.0,
        'ls_on' : 10,
    }),
}

options = {
    'nls' : 'newton',
    'ls' : 'ls',
    'save_times' : 'all',
}



def test_mesh_expand():
    import numpy as nm
    from sfepy.mesh.mesh_tools import expand2d
    from sfepy.discrete.fem.mesh import Mesh
    from sfepy import data_dir

    mesh2d = Mesh.from_file(data_dir + '/meshes/2d/square_quad.mesh')
    mesh3d_ref = Mesh.from_file(data_dir +\
                                '/meshes/3d/cube_medium_hexa.mesh')
    mesh3d_gen = expand2d(mesh2d, 0.1, 10)
    mesh3d_gen.coors[:,2] += -0.5

    d0 = nm.sort(nm.linalg.norm(mesh3d_ref.coors, axis=1))
    d1 = nm.sort(nm.linalg.norm(mesh3d_gen.coors, axis=1))
    ok =  nm.linalg.norm(d0 - d1) < 1e-6
    assert ok



import numpy as nm
import scipy.sparse as sps
import pytest

import sfepy.base.testing as tst

@pytest.fixture(scope='module')
def data():
    from sfepy.base.base import Struct
    conn = nm.array([[0, 1, 2],
                     [2, 3, 4]], dtype=nm.int32)

    num = conn.max() + 1

    iels = nm.array([0, 1], dtype=nm.int32)

    vec_in_els = nm.array([[1, 1, 1], [2, 2, 2]], dtype=nm.float64)
    vec_in_els.shape = (2, 1, 1, 3)

    mtx_in_els = nm.array([nm.ones((3, 3)), 2 * nm.ones((3, 3))],
                          dtype=nm.float64)
    mtx_in_els.shape = (2, 1, 3, 3)

    return Struct(conn=conn, num=num, iels=iels,
                  vec_in_els=vec_in_els, mtx_in_els=mtx_in_els)

def test_assemble_vector(data):
    from sfepy.discrete.common.extmods.assemble import assemble_vector

    vec = nm.zeros(data.num, dtype=nm.float64)

    assemble_vector(vec, data.vec_in_els, data.iels, 1, data.conn)

    aux = nm.array([1, 1, 3, 2, 2], dtype=nm.float64)

    tst.report('assembled: %s' % vec)
    tst.report('expected: %s' % aux)
    ok = tst.compare_vectors(vec, aux, label1='assembled', label2='expected')
    assert ok

def test_assemble_vector_complex(data):
    from sfepy.discrete.common.extmods.assemble import assemble_vector_complex

    vec = nm.zeros(data.num, dtype=nm.complex128)
    vec_in_els = data.vec_in_els.astype(nm.complex128) * (2 - 3j)

    assemble_vector_complex(vec, vec_in_els, data.iels, 1, data.conn)

    aux = nm.array([2-3j, 2-3j, 6-9j, 4-6j, 4-6j],
                   dtype=nm.complex128)

    tst.report('assembled: %s' % vec)
    tst.report('expected: %s' % aux)
    ok = tst.compare_vectors(vec, aux, label1='assembled', label2='expected')
    assert ok

def test_assemble_matrix(data):
    from sfepy.discrete.common.extmods.assemble import assemble_matrix

    mtx = sps.csr_matrix(nm.ones((data.num, data.num),
                                 dtype=nm.float64))
    mtx.data[:] = 0.0

    assemble_matrix(mtx.data, mtx.indptr, mtx.indices, data.mtx_in_els,
                    data.iels, 1, data.conn, data.conn)

    aux = nm.array([[1, 1, 1, 0, 0],
                    [1, 1, 1, 0, 0],
                    [1, 1, 3, 2, 2],
                    [0, 0, 2, 2, 2],
                    [0, 0, 2, 2, 2]], dtype=nm.float64)

    tst.report('assembled:\n%s' % mtx.toarray())
    tst.report('expected:\n%s' % aux)
    ok = tst.compare_vectors(mtx, aux, label1='assembled', label2='expected')
    assert ok

def test_assemble_matrix_complex(data):
    from sfepy.discrete.common.extmods.assemble import assemble_matrix_complex

    mtx = sps.csr_matrix(nm.ones((data.num, data.num),
                                 dtype=nm.complex128))
    mtx.data[:] = 0.0
    mtx_in_els = data.mtx_in_els.astype(nm.complex128) * (2 - 3j)

    assemble_matrix_complex(mtx.data, mtx.indptr, mtx.indices, mtx_in_els,
                            data.iels, 1, data.conn, data.conn)

    aux = nm.array([[2-3j, 2-3j, 2-3j, 0+0j, 0+0j],
                    [2-3j, 2-3j, 2-3j, 0+0j, 0+0j],
                    [2-3j, 2-3j, 6-9j, 4-6j, 4-6j],
                    [0+0j, 0+0j, 4-6j, 4-6j, 4-6j],
                    [0+0j, 0+0j, 4-6j, 4-6j, 4-6j]], dtype=nm.complex128)

    tst.report('assembled:\n%s' % mtx.toarray())
    tst.report('expected:\n%s' % aux)
    ok = tst.compare_vectors(mtx, aux, label1='assembled', label2='expected')
    assert ok



import numpy as nm

from sfepy.homogenization.engine import HomogenizationEngine as he
from sfepy.homogenization.engine import HomogenizationWorkerMulti as hwm
import sfepy.base.testing as tst

def test_dependencies():
    get_deps = hwm.get_sorted_dependencies

    coefs = {'A' : {'requires' : ['a', 'd', 'c.B']},
             'B' : {'requires' : ['b']}}
    requirements = {'a' : {'requires' : ['b', 'c']},
                    'b' : {'requires' : ['c']},
                    'c' : {},
                    'd' : {'requires' : ['b', 'a']}}

    deps = get_deps(requirements, coefs, None)
    ok = ((deps == ['c', 'b', 'a', 'd', 'c.B', 'c.A'])
          or (deps == ['c', 'b', 'c.B', 'a', 'd', 'c.A'])
          or (deps == ['c', 'b', 'a', 'c.B', 'd', 'c.A']))
    tst.report(deps, ':', ok)

    coefs['B']['requires'] = ['b', 'c.A']

    try:
        deps = get_deps(requirements, coefs, None)

    except ValueError as err:
        tst.report('detected:', str(err))
        _ok = 'circular requirement "c.' in str(err)

    else:
        _ok = False

    tst.report('circular dependency detection 1:', _ok)
    ok = ok and _ok

    coefs['B']['requires'] = ['b']
    requirements['c']['requires'] = ['d']

    try:
        deps = get_deps(requirements, coefs, None)

    except ValueError as err:
        tst.report('detected:', str(err))
        _ok = 'circular requirement' in str(err)

    else:
        _ok = False

    tst.report('circular dependency detection 2:', _ok)
    ok = ok and _ok

    assert ok

def test_chunk_micro():
    coefs = {'A' : {'requires' : ['a', 'd', 'c.B']},
             'B' : {'requires' : ['b']}}
    requirements = {'a' : {'requires' : ['b', 'c']},
                    'b' : {'requires' : ['c']},
                    'c' : {},
                    'd' : {'requires' : ['b', 'a']}}

    volumes = {'total': {'expression': ''}}
    coefs = he.define_volume_coef(coefs, volumes)
    orig_deps_num = len(requirements) + len(coefs)

    num_workers, num_micro, chunks_per_worker = 5, 61, 2
    store_micro_idxs = [0, 1, 18, 20, 21]
    micro_chunk_tab, requirements, coefs = \
        hwm.chunk_micro_tasks(num_workers, num_micro, requirements, coefs,
                              chunks_per_worker, store_micro_idxs)

    dep_names = hwm.get_sorted_dependencies(requirements, coefs, None)

    ok = (orig_deps_num * len(micro_chunk_tab)) == len(dep_names)
    tst.report('splitting into chunks:', ok)

    deps = {}
    for k in dep_names:
        chunk_id = int(k[-3:])
        nmic = len(range(*micro_chunk_tab[chunk_id].indices(num_micro)))
        deps[k] = [1] * nmic
        if k[2:] in coefs and 'Volume_total' not in k:
            reqs = '#'.join(coefs[k[2:]]['requires'])
            ok = ok and 'Volume_total' in reqs

    tst.report('volume dependency:', ok)

    deps = hwm.dechunk_reqs_coefs(deps, len(micro_chunk_tab))

    ok = ok and\
        nm.all([(nm.sum(v) == num_micro) for v in deps.values()])
    tst.report('merging chunks:', ok)

    assert ok



import os.path as op
import inspect

import numpy as nm
import numpy.linalg as nla

from sfepy.base.base import Struct, assert_, IndexedStruct

def report(*argc):
    """All tests should print via this function."""
    format = '...' + ' %s' * len(argc)
    msg = format % argc
    print(msg)

def eval_coor_expression(expression, coor):

    x = coor[:, 0]
    y = coor[:, 1]
    if coor.shape[1] == 3:
        z = coor[:, 2]
    else:
        z = None

    env = {'x' : x, 'y' : y, 'z' : z}
    out = eval(expression, nm.__dict__, env)

    if isinstance(out, float):
        aux = nm.empty(coor.shape[0], dtype=nm.float64)
        aux.fill(out)
        out = aux

    return out

def compare_vectors(vec1, vec2, allowed_error=1e-8,
                    label1='vec1', label2='vec2', norm=None):

    diff_norm = nla.norm(vec1 - vec2, ord=norm)
    report('||%s - %s||: %e' % (label1, label2, diff_norm))
    if diff_norm > allowed_error:
        return False
    else:
        return True

def assert_equal(a, b, msg='assertion of equality failed!'):
    import scipy.sparse

    assert_base_types = (int, float, str, bytes, complex,
                         None.__class__, type, nm.number)
    if a is b: return

    def assert_dict(a, b):
        assert_(set(a.keys()) == set(b.keys()), msg)
        for i in a:
            assert_equal(a[i], b[i], msg)

    def assert_list(a, b):
        assert_(len(a) == len(b), msg)
        for i, j in zip(a, b):
            assert_equal(i, j)

    assert_(a.__class__ is b.__class__, msg)
    if isinstance(a, (int, float, str, assert_base_types, bytes, complex)):
        assert_(a == b, msg)

    elif isinstance(a, dict):
        assert_dict(a, b)

    elif isinstance(a, (list, tuple)):
        assert_list(a, b)

    elif isinstance(a, nm.ndarray):
        nm.testing.assert_array_equal(a,b)

    elif isinstance(a, (scipy.sparse.csr_matrix, scipy.sparse.csc_matrix)):
        nm.testing.assert_array_equal(a.data, b.data)
        nm.testing.assert_array_equal(a.indices, b.indices)
        nm.testing.assert_array_equal(a.indptr, b.indptr)

    elif isinstance(a, object):
        cls = a.__class__

        if hasattr(cls, '__slots__'):
            ad = dict((i,getattr(a,i)) for i in cls.__slots__)
            bd = dict((i,getattr(b,i)) for i in cls.__slots__)

        elif hasattr(a, '__dict__'):
            ad = a.__dict__
            bd = b.__dict__

        else:
            def members(obj):
                out = inspect.getmembers(obj, lambda x: not
                                         inspect.isroutine(x) )
                out = dict( (k,v) for k,v in out if not k.startswith('__'))
                return out
            ad = members(a)
            bd = members(b)

        assert_dict(ad, bd)

class NLSStatus(IndexedStruct):
    """
    Custom nonlinear solver status storing stopping condition of all
    time steps.
    """
    def __setitem__(self, key, val):
        IndexedStruct.__setitem__(self, key, val)
        if key == 'condition':
            self.conditions.append(val)

def check_conditions(conditions):
    ok = (conditions == 0).all()
    if not ok:
        report('nls stopping conditions:')
        report(conditions)
    return ok

def run_declaratice_example(ex_filename, output_dir, ext='.vtk',
                            remove_prefix=''):
    """
    Run a declarative example in `ex_filename` given relatively to
    ``sfepy.base_dir``.
    """
    import sfepy
    from sfepy.applications import solve_pde

    report('solving %s...' % ex_filename)

    if remove_prefix and ex_filename.startswith(remove_prefix):
        output_name = ex_filename.replace(remove_prefix, '')

    else:
        output_name = ex_filename

    output_name = op.splitext(output_name.replace('/', '-'))[0]
    filename = op.join(sfepy.base_dir, ex_filename)
    name = op.splitext(op.split(output_name)[1])[0]
    fmt = ext.replace('.', '')
    options = Struct(output_filename_trunk=name,
                     output_format=fmt if fmt != '' else 'vtk',
                     save_ebc=False, save_ebc_nodes=False,
                     save_regions=False,
                     save_regions_as_groups=False,
                     solve_not=False)
    status = IndexedStruct(nls_status=NLSStatus(conditions=[]))

    solve_pde(filename, options=options, status=status, output_dir=output_dir)
    report('%s solved' % ex_filename)

    return nm.array(status.nls_status.conditions)



import pytest

import sfepy.base.testing as tst

input_names = {'TL': 'examples/large_deformation/hyperelastic.py',
               'UL': 'examples/large_deformation/hyperelastic_ul.py',
               'ULM': 'examples/large_deformation/hyperelastic_ul_up.py'}
output_name_trunk = 'test_hyperelastic_'

@pytest.mark.slow
def test_solution(output_dir):
    import sfepy
    from sfepy.base.base import IndexedStruct, Struct
    from sfepy.applications import solve_pde
    import numpy as nm
    import os.path as op

    solutions = {}
    ok = True

    for hp, pb_filename in input_names.items():

        input_name = op.join(sfepy.base_dir, pb_filename)

        name = output_name_trunk + hp
        solver_options = Struct(output_filename_trunk=name,
                                output_format='vtk',
                                save_ebc=False, save_ebc_nodes=False,
                                save_regions=False,
                                save_regions_as_groups=False,
                                solve_not=False)

        tst.report('hyperelastic formulation: %s' % hp)

        status = IndexedStruct(nls_status=tst.NLSStatus(conditions=[]))

        pb, state = solve_pde(input_name, solver_options, status=status,
                              output_dir=output_dir)
        converged = status.nls_status.condition == 0
        ok = ok and converged

        solutions[hp] = state.get_state_parts()['u']
        tst.report('%s solved' % input_name)

    rerr = 1.0e-3
    aerr = nm.linalg.norm(solutions['TL'], ord=None) * rerr

    tst.report('allowed error: rel = %e, abs = %e' % (rerr, aerr))
    ok = ok and tst.compare_vectors(solutions['TL'], solutions['UL'],
                                    label1='TLF',
                                    label2='ULF',
                                    allowed_error=rerr)

    ok = ok and tst.compare_vectors(solutions['UL'], solutions['ULM'],
                                    label1='ULF',
                                    label2='ULF_mixed',
                                    allowed_error=rerr)

    assert ok



import pytest

def pytest_configure(config):
    config.addinivalue_line(
        'markers',
        'slow: Mark tests as slow (deselect with \'-m "not slow"\'.',
    )

def pytest_addoption(parser):
    parser.addoption('--output-dir', action='store', default=None)

@pytest.fixture(scope='session')
def output_dir(request, tmpdir_factory):
    """
    Output directory for tests.
    """
    output_dir = request.config.getoption('output_dir')
    if output_dir is not None:
        import os
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        if not os.path.isdir(output_dir):
            raise IOError(f'cannot create directory "{output_dir}"!')

        return output_dir

    else:
        return tmpdir_factory.mktemp('output')



"""
Modal analysis of a wheel set.

The first six modes are the rigid body modes because no boundary
conditions are applied.

Running the simulation::

  sfepy-run sfepy/examples/linear_elasticity/modal_analysis_declarative.py

The eigenvalues are saved to wheelset_eigs.txt and the eigenvectros to
wheelset.vtk. View the results using::

  sfepy-view wheelset.vtk -f u003:wu003:f30%:p0 1:vw:p0

The first six frequencies calculated by SfePy::

  [11.272, 11.322, 34.432, 80.711, 80.895, 93.149]

The results of modal analysis performed in Ansys::

  [11.306, 11.316, 34.486, 80.901, 81.139, 93.472]
"""
import numpy as nm
from sfepy.base.base import output
from sfepy.mechanics.matcoefs import stiffness_from_youngpoisson
from sfepy import data_dir


def report_eigs(pb, evp):
    eigs = evp.eigs
    n_rbm = evp.pb.conf.n_rbm

    output('%d eigenvalues converged (%d ignored as rigid body modes)' %
           (len(eigs), n_rbm))

    eigs = eigs[n_rbm:]

    omegas = nm.sqrt(eigs)
    freqs = omegas / (2 * nm.pi)

    output('number |         eigenvalue |  angular frequency '
           '|          frequency')
    for ii, eig in enumerate(eigs):
        output('%6d | %17.12e | %17.12e | %17.12e'
               % (ii + 1, eig, omegas[ii], freqs[ii]))


def define(n_eigs=6, approx_order=1, density=7850., young=210e9, poisson=0.3):
    filename_mesh, dim = data_dir + '/meshes/3d/wheelset.vtk', 3

    n_rbm = 0

    options = {
        'n_eigs': n_eigs + n_rbm,
        'eigs_only': False,
        'post_process_hook_final': 'report_eigs',
        'evps': 'eig',
    }

    regions = {
        'Omega': 'all',
        'Fix': ('vertices in (x < -1.08999)', 'vertex'),
    }

    materials = {
        'm': ({
            'D': stiffness_from_youngpoisson(dim, young, poisson),
            'rho': density,
        },),
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', approx_order),
    }

    variables = {
        'u': ('unknown field', 'displacement'),
        'v': ('test field', 'displacement', 'u'),
    }

    integrals = {
        'i': 2 * approx_order,
    }

    equations = {
        'lhs': 'dw_lin_elastic.i.Omega(m.D, v, u)',
        'rhs': 'dw_dot.i.Omega(m.rho, v, u)',
    }

    ebcs = {
        'fix': ('Fix', {'u.all': 0.0})  # fix rigid body modes
    }

    solvers = {
        # 'eig': ('eig.matlab', {
        #     'method': 'eigs',
        #     'which': 'sm',
        #     'eps': 1e-6,
        # }),
        #'eig': ('eig.primme', {
        #    'which': 'SM',
        #    'tol': 1e-8,
        #}),
        'eig': ('eig.scipy', {
            'method': 'eigsh',
            'which': 'LM',
            'sigma': 0,
            'tol': 1e-8,
            'maxiter': 1000,
        }),
    }

    return locals()



#!/usr/bin/env python
"""
Generate available fields table for ReST documentation.
"""
import os.path as op
import sys

from argparse import ArgumentParser

sys.path.append('.')

import sfepy
from sfepy import get_paths
from sfepy.base.base import load_classes
from sfepy.discrete import Field

field_files = [ii for ii
               in get_paths('sfepy/discrete/fem/fields*.py')
               if 'fields_base.py' not in ii]
field_files += get_paths('sfepy/discrete/structural/fields*.py')
field_files += get_paths('sfepy/discrete/iga/fields*.py')
field_files += get_paths('sfepy/discrete/dg/fields.py')
field_table = load_classes(field_files, [Field], ignore_errors=True,
                           name_attr='family_name')


header = """
.. tabularcolumns:: |l|l|l|p{0.55\linewidth}|
.. list-table:: Fields
   :widths: 5 15 15 65
   :header-rows: 1

   * - space
     - basis
     - region kind
     - description
"""

table_row = """   * - %s
     - %s
     - %s
     - %s
"""

class_link = ':class:`{short} <{full}>`'

translate_rtype = {'volume' : 'cell', 'surface' : 'facet'}

def typeset_field_table(fd, field_table):

    fd.write('.. _field_table:\n')
    fd.write(header)

    rows = {}
    for key, cls in field_table.items():
        rtype, space, *basis = key.split('_')
        rtype = translate_rtype[rtype]
        basis = '_'.join(basis)
        doc = cls.__doc__
        if doc is not None:
            desc = doc.strip().splitlines()[0]

        else:
            desc = ''

        sdata = rows.setdefault(space, {})
        row = sdata.setdefault(basis, ([], desc))

        name = cls.__module__ + '.' + cls.__name__
        row[0].append(class_link.format(short=rtype, full=name))
        if row[1] is None:
            row[1] = desc

    for space, sdata in rows.items():
        for basis, (rtypes, desc) in sorted(sdata.items(), key=lambda x: x[0]):
            rtype = ', '.join(rtypes)
            fd.write(table_row % (space, basis, rtype, desc))

    fd.write('\n')


def typeset(fd):
    """
    Utility function called by Sphinx.
    """
    fd = open(fd, 'w')
    typeset_field_table(fd, field_table)
    fd.close()


def gen_field_table(app):
    typeset(op.join(app.builder.srcdir, 'field_table.rst'))


def setup(app):
    app.connect('builder-inited', gen_field_table)


def main():
    parser = ArgumentParser(description=__doc__)
    parser.add_argument('-v', '--version', action='version',
                        version='%(prog)s ' + sfepy.__version__)
    parser.add_argument('-o', '--output', metavar='output_filename',
                        action='store', dest='output_filename',
                        default='field_table.rst')
    options = parser.parse_args()

    typeset(options.output_filename)


if __name__ == '__main__':
    main()



import sfepy.base.testing as tst

def test_parse_equations():
    from sfepy.discrete.parse_equations import create_bnf

    test_strs = [
        """- ev_volume.i1.Omega(uc)""",
        """- 2 * dw_term.i1.Omega(uc)
         = - 3.0 * dw_term2.i1.Omega2(uc)""",
        """2 * dw_term.i1.Omega(uc)
         = 3.0 * dw_term2.i1.Omega2(uc)""",
        """- (2 + 1j - 3) * dw_term.i1.Omega(uc)
           = - (1j - 3.0 + 2.0j) * dw_term2.i1.Omega2(uc)""",
        """(2 + 1j) * dw_term.i1.Omega(uc)
           = (3.0 + 2.0j) * dw_term2.i1.Omega2(uc)""",
        """d_term1.Y(fluid, u, w, Nu, dcf, mode)
             + d_term2.Omega(u, w, Nu, dcf, mode)
             - d_another_term.Elsewhere(w, p, Nu, dcf, mode)
             = - dw_rhs.a.Y3(u, q, Nu, dcf, mode)""",
        """no_args() = 0""",
        """+ something(a, b, c) = + something_else(c, a, d[-1])""",
        """term_.a.a(u)""",
        """term.i1.Omega(v, du/dt) + term2.i2.Gamma(v, dphi/dt)""",
        """dw_jump.isurf.Gamma12_1(jump1.val, q1, p1, tr(p2))""",
        """term_with_trace.2.Gamma1(v, tr(Gamma2, u))""",
    ]

    n_fail = 0
    term_descs = []
    for test_str in test_strs:
        term_descs[:] = []
        try:
            bnf = create_bnf(term_descs)
            bnf.parseString(test_str)
        except:
            tst.report('failed: %s' % test_str)
            n_fail += 1

    tst.report('%d failure(s)' % n_fail)

    assert n_fail == 0

def test_parse_regions():
    from sfepy.discrete.parse_regions import create_bnf

    test_strs = ['vertices of surface -v r.Omega',
                 'r.Y_2 +v copy r.Y_1',
                 'vertices in (y <= 0.00001) & (x < 0.11)',
                 'vertices in ((y <= 0.00001) & (x < 0.11))',
                 'vertices in (((y <= 0.00001) & (x < 0.11)))',
                 'vertices in (((0.00001 < y) & (x < 0.11)))',
                 'vertices of group 0',
                 """vertices of group 0 +v vertices of group 1
                 +c cells by afun""",
                 'all -v vertices in (y == 0.00001)',
                 'all -v vertices of surface',
                 'all -c r.DOmega_100',
                 """r.Y_1 -v vertices of surface *c r.Z_8
                    *v vertices in (y > 0)""",
                 'vertices of surface +v vertices by pokus',
                 'cells of group 6 +c vertices by fn2_3c',
                 """r.Y_1 *v (r.Y_2 +c (vertices in (y > 0) *v r.Y_32))
                    -v vertices of surface -c r.Y_5""",
                 'copy r.ab2-b-c +v r.d12-23',
                 'vertices by afun',
                 'vertex in r.Gamma_3',
                 'vertex 10',
                 'vertex 10, 20, 30',
                 'cell 10',
                 'cell 10, 20, 30',
                 'vertex 10, 20 +v cell 30, 40',
                 '(vertex 10, 20) +v (cell 30, 40)',
                 'cell 10, 20, 30 +v vertex 10',
                 'cells by afun']

    stack = []
    bnf = create_bnf(stack)

    n_fail = 0
    for test_str in test_strs:
        stack[:] = []

        try:
            bnf.parseString(test_str)

        except:
            tst.report('failed: %s' % test_str)
            n_fail += 1

    tst.report('%d failures' % n_fail)

    assert n_fail == 0



import os.path as op

import numpy as nm

import sfepy
from sfepy.discrete.common import Field
import sfepy.discrete.common.global_interp as gi
import sfepy.base.testing as tst

def test_ref_coors_fem():
    from sfepy.discrete.fem import Mesh, FEDomain

    mesh = Mesh.from_file('meshes/3d/special/cross3d.mesh',
                          prefix_dir=sfepy.data_dir)
    domain = FEDomain('domain', mesh)

    omega = domain.create_region('Omega', 'all')

    field = Field.from_args('linear', nm.float64, 'scalar', omega,
                            approx_order=1)


    mcoors = field.domain.get_mesh_coors()
    conn = field.domain.get_conn()

    bbox = field.domain.get_mesh_bounding_box()
    ray = nm.linspace(bbox[0, 0], bbox[1, 0], 7)
    coors = nm.zeros((ray.shape[0], 3), dtype=nm.float64)

    def gen_rays():
        coors[:, 0] = ray
        yield coors

        coors.fill(0.0)
        coors[:, 1] = ray
        yield coors

        coors.fill(0.0)
        coors[:, 2] = ray
        yield coors

    ok = True

    ctx = field.create_basis_context()._geo_ctx

    for ir, coors in enumerate(gen_rays()):
        tst.report('ray %d' % ir)
        ref_coors, cells, status = gi.get_ref_coors(field, coors,
                                                    strategy='general',
                                                    close_limit=0.0,
                                                    verbose=False)
        tst.report(ref_coors)
        tst.report(cells)
        tst.report(status)

        # In the distorted cell 2, the Newton method finds a solution
        # outside of the cell. This will be fixed when box constraints
        # are applied.
        _ok = nm.all((status == 0) | ((cells == 2) & (status == 3)))
        if not _ok:
            tst.report('wrong status %s for ray %d!' % (status, ir))

        ok = ok and _ok

        for ic, cell in enumerate(cells):
            ctx.iel = cell
            bf = ctx.evaluate(ref_coors[ic:ic+1], check_errors=False)

            cell_coors = mcoors[conn[cell]]
            coor = nm.dot(bf, cell_coors).ravel()

            _ok = nm.allclose(coor, coors[ic], atol=1e-14, rtol=0.0)
            if not _ok:
                tst.report('ray %d point %d:' % (ir, ic))
                tst.report(' - wrong reference coordinates %s!'
                           % ref_coors[ic])
                tst.report(' - given point: %s' % coors[ic])
                tst.report(' - found point: %s' % coor)
            ok = ok and _ok

    assert ok

def test_ref_coors_iga():
    from sfepy.discrete.iga.domain import IGDomain

    domain = IGDomain.from_file(op.join(sfepy.data_dir,
                                        'meshes/iga/block2d.iga'))

    omega = domain.create_region('Omega', 'all')

    field = Field.from_args('iga', nm.float64, 'scalar', omega,
                            approx_order='iga', poly_space_base='iga')

    mcoors = field.nurbs.cps
    conn = field.get_econn('cell', field.region)

    bbox = domain.eval_mesh.get_bounding_box()
    ray = nm.linspace(bbox[0, 0], bbox[1, 0], 11)
    coors = nm.c_[ray, ray]

    ref_coors, cells, status = gi.get_ref_coors(field, coors,
                                                strategy='general',
                                                close_limit=0.0,
                                                verbose=False)
    tst.report(ref_coors)
    tst.report(cells)
    tst.report(status)

    ok = nm.all(status == 0)

    ctx = field.create_basis_context()

    for ic, cell in enumerate(cells):
        ctx.iel = cell
        bf = ctx.evaluate(ref_coors[ic:ic+1])

        cell_coors = mcoors[conn[cell]]
        coor = nm.dot(bf, cell_coors).ravel()

        _ok = nm.allclose(coor, coors[ic], atol=1e-14, rtol=0.0)
        if not _ok:
            tst.report('point %d:' % ic)
            tst.report(' - wrong reference coordinates %s!'
                       % ref_coors[ic])
            tst.report(' - given point: %s' % coors[ic])
            tst.report(' - found point: %s' % coor)
        ok = ok and _ok

    assert ok



from functools import partial

import numpy as nm
import pytest

from sfepy.discrete.fem.meshio import UserMeshIO
from sfepy.mesh.mesh_generators import gen_block_mesh
import sfepy.mechanics.matcoefs as mc
import sfepy.base.testing as tst

def define(t1=15e-6, dt=1e-6, dims=(0.1, 0.02, 0.005), shape=(11, 3, 3),
           young=70e9, poisson=0.3, density=2700, mass_lumping='row_sum',
           mass_beta=0.2):

    def mesh_hook(mesh, mode):
        """
        Generate the block mesh.
        """
        if mode == 'read':
            mesh = gen_block_mesh(dims, shape, 0.5 * nm.array(dims),
                                  name='user_block', verbose=False)
            return mesh

        elif mode == 'write':
            pass

    filename_mesh = UserMeshIO(mesh_hook)
    dim = len(dims)

    def get_sensor(coors, domain=None):
        ii = coors.argmax(axis=0)
        return ii[:1]

    regions = {
        'Omega' : 'all',
        'Left' : ('vertices in (x < 1e-12)', 'facet'),
        'Sensor' : ('vertices by get_sensor', 'vertex'),
    }

    materials = {
        'solid' : ({
            'D': mc.stiffness_from_youngpoisson(
                dim=dim, young=young, poisson=poisson, plane='strain'
            ),
            'rho': density,
            '.lumping' : mass_lumping,
            '.beta' : mass_beta,
         },),
    }

    fields = {
        'displacement': ('real', 'vector', 'Omega', 2),
    }

    integrals = {
        'i' : 4,
    }

    variables = {
        'u' : ('unknown field', 'displacement', 0),
        'du' : ('unknown field', 'displacement', 1),
        'ddu' : ('unknown field', 'displacement', 2),
        'v' : ('test field', 'displacement', 'u'),
        'dv' : ('test field', 'displacement', 'du'),
        'ddv' : ('test field', 'displacement', 'ddu'),
    }
    var_names = {'u' : 'u', 'du' : 'du', 'ddu' : 'ddu'}

    ebcs = {
        'fix' : ('Left', {'u.all' : 0.0, 'du.all' : 0.0, 'ddu.all' : 0.0}),
    }

    def get_ic(coors, ic, mode='u'):
        val = nm.zeros_like(coors)
        if mode == 'u':
            val[:, 0] = 0.0

        elif mode == 'du':
            xmax = coors[:, 0].max()
            val[:, -1] = nm.where((coors[:, 0] > (xmax - 1e-12)), 1.0, 0.0)

        return val

    functions = {
        'get_sensor' : (get_sensor,),
        'get_ic_u' : (get_ic,),
        'get_ic_du' : (lambda coor, ic: get_ic(coor, None, mode='du'),),
    }

    ics = {
        'ic' : ('Omega', {'u.all' : 'get_ic_u', 'du.all' : 'get_ic_du'}),
    }

    equations = {
        'balance_of_forces' :
        """dw_dot.i.Omega(solid.rho, ddv, ddu)
         + dw_zero.i.Omega(dv, du)
         + dw_lin_elastic.i.Omega(solid.D, v, u) = 0""",
    }

    solvers = {
        'ls' : ('ls.auto_direct', {
            # Reuse the factorized linear system from the first time step.
            'use_presolve' : True,
            # Speed up the above by omitting the matrix digest check used
            # normally for verification that the current matrix corresponds to
            # the factorized matrix stored in the solver instance. Use with
            # care!
            'use_mtx_digest' : False,
        }),
        'lsrmm' : ('ls.rmm', {
            'rmm_term' : """de_mass.i.Omega(solid.rho, solid.lumping,
                                            solid.beta, ddv, ddu)""",
            'debug' : False,
        }),
        'newton' : ('nls.newton', {
            'i_max'      : 1,
            'eps_a'      : 1e-6,
            'eps_r'      : 1e-6,
        }),
        'tsvv' : ('ts.velocity_verlet', {
            # Explicit method.
            't0' : 0.0,
            't1' : t1,
            'dt' : 0.1 * dt,
            'n_step' : None,

            'is_linear'  : True,

            'var_names' : var_names,
            'verbose' : 1,
        }),
        'tscd' : ('ts.central_difference', {
            # Explicit method.
            't0' : 0.0,
            't1' : t1,
            'dt' : 0.1 * dt,
            'n_step' : None,

            'is_linear'  : True,

            'var_names' : var_names,
            'verbose' : 1,
        }),
        'tsn' : ('ts.newmark', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'beta' : 0.25,
            'gamma' : 0.5,

            'var_names' : var_names,
            'verbose' : 1,
        }),
        'tsga' : ('ts.generalized_alpha', {
            't0' : 0.0,
            't1' : t1,
            'dt' : dt,
            'n_step' : None,

            'is_linear'  : True,

            'rho_inf' : 0.95,
            'alpha_m' : None,
            'alpha_f' : None,
            'beta' : None,
            'gamma' : None,

            'var_names' : var_names,
            'verbose' : 1,
        }),
        'tsb' : ('ts.bathe', {
            't0' : 0.0,
            't1' : t1,
            'dt' : 0.5 * dt,
            'n_step' : None,

            'is_linear'  : True,

            'var_names' : var_names,
            'verbose' : 1,
        }),
        'tscedb' : ('tsc.ed_basic', {
            'eps_r' : (1e-3, 1e-1),
            'eps_a' : (1e-6, 1e-2),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
        }),
        'tscedl' : ('tsc.ed_linear', {
            'eps_r' : (1e-3, 1e-1),
            'eps_a' : (1e-6, 1e-2),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
            'fred' : 0.9,
            'inc_wait' : 10,
            'min_finc' : 1.7,
        }),
        'tscedpid' : ('tsc.ed_pid', {
            'eps_r' : (1e-3, 1e-1),
            'eps_a' : (1e-6, 1e-2),
            'fmin' : 0.3,
            'fmax' : 2.5,
            'fsafety' : 0.85,
            'pcoef' : 0.4,
            'icoef' : 0.3,
            'dcoef' : 0,
        }),
    }

    options = {
        'ts' : 'tsn',
        'nls' : 'newton',
        'ls' : 'ls',

        'save_times' : 31,

        'active_only' : False,

        'output_format' : 'h5',
    }

    return locals()

@pytest.fixture(scope='module')
def problem():
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    define_dict = define()
    conf = ProblemConf.from_dict(define_dict, sys.modules[__name__])

    pb = Problem.from_conf(conf)
    pb.update_materials()

    # Get full size matrices for energy calculations.
    pb.init_solvers()
    tss = pb.solver
    ebcs = pb.conf.ebcs
    pb.time_update(ebcs={})
    tss.constant_matrices = None
    pb.Mf, Cf, pb.Kf = tss.get_matrices(tss.nls, pb.set_default_state()())

    # Restore EBCs.
    pb.time_update(ebcs=ebcs)

    return pb

def _list_solvers(confs, kind='ts'):
    d = [val for val in confs.values() if val.kind.startswith(kind+'.')]
    d.sort(key=lambda a: a.name)

    return d

@pytest.mark.slow
def test_ed_solvers(problem, output_dir):
    from scipy.integrate import simpson
    from sfepy.base.base import IndexedStruct

    tss_confs = _list_solvers(problem.solver_confs)
    tsc_confs = _list_solvers(problem.solver_confs, kind='tsc')

    vu = problem.get_variables()['u']
    sensor = problem.domain.regions['Sensor']
    isens = 3 * vu.field.get_dofs_in_region(sensor)[0] + 2

    def store_ths(pb, ts, variables, ths):
        sp = variables.get_state_parts()
        u1, v1, a1 = sp['u'], sp['du'], sp['ddu']

        e_u = 0.5 * u1 @ pb.Kf @ u1
        e_t = 0.5 * v1 @ pb.Mf @ v1
        ths.append((ts.time, u1[isens], v1[isens], a1[isens],
                    e_u, e_t, e_u + e_t))

    all_ths = []
    stats = []
    t1s = []
    for tsc_conf in [None] + tsc_confs:
        problem.tsc_conf = None
        for tss_conf in tss_confs:
            status = IndexedStruct()
            problem.init_solvers(tsc_conf=tsc_conf, ts_conf=tss_conf,
                                 status=status, force=True)
            ths = []
            problem.solve(status=status, save_results=False,
                          step_hook=partial(store_ths, ths=ths))
            all_ths.append(nm.array(ths))
            stats.append((problem.solver.tsc.conf.kind, tss_conf.kind,
                          status.n_step, status.time))
            t1s.append(problem.solver.ts.time)

    kinds = [val[0:2] for val in stats]
    stats.sort(key=lambda x: x[-1])
    tst.report('solution times / numbers of time steps:')
    for row in stats:
        tst.report('%.2f [s] / % 4d' % (row[3], row[2]), ':', row[:2])

    # import matplotlib.pyplot as plt
    # for ii, ths in enumerate(all_ths):
    #     fig, ax = plt.subplots()
    #     ax.plot(ths[:,0], ths[:,4])
    #     ax.plot(ths[:,0], ths[:,5])
    #     ax.plot(ths[:,0], ths[:,6])
    #     ax.set_title(kinds[ii])
    # plt.show()

    all_iths = nm.array(
        [[simpson(val, ths[:, 0]) for val in ths.T[1:]] for ths in all_ths]
    )

    tst.report('status, solver: time integral of (u, v, a, e_u, e_t, e_u-e_t)')
    iths_ref = all_iths[0]
    e0 = all_ths[0][0, -1]
    e_rtols = {
        ('tsc.fixed', 'ts.bathe') : 1e-1,
        ('tsc.fixed', 'ts.generalized_alpha') : 1e-2,
        ('tsc.fixed', 'ts.newmark') : 1e-12,
        ('tsc.fixed', 'ts.central_difference') : 1e-2,
        ('tsc.fixed', 'ts.velocity_verlet') : 1e-2,
        ('tsc.ed_basic', 'ts.bathe') : 1e-1,
        ('tsc.ed_basic', 'ts.generalized_alpha') : 1e-3,
        ('tsc.ed_basic', 'ts.newmark') : 1e-12,
        ('tsc.ed_basic', 'ts.central_difference') : 2e-2,
        ('tsc.ed_basic', 'ts.velocity_verlet') : 2e-2,
        ('tsc.ed_linear', 'ts.bathe') : 1e-1,
        ('tsc.ed_linear', 'ts.generalized_alpha') : 1e-3,
        ('tsc.ed_linear', 'ts.newmark') : 1e-12,
        ('tsc.ed_linear', 'ts.central_difference') : 2e-2,
        ('tsc.ed_linear', 'ts.velocity_verlet') : 2e-2,
        ('tsc.ed_pid', 'ts.bathe') : 1e-2,
        ('tsc.ed_pid', 'ts.generalized_alpha') : 1e-4,
        ('tsc.ed_pid', 'ts.newmark') : 1e-12,
        ('tsc.ed_pid', 'ts.central_difference') : 1e-2,
        ('tsc.ed_pid', 'ts.velocity_verlet') : 1e-2,
    }
    ok = True
    for ii, iths in enumerate(all_iths):
        ienergy = e0 * t1s[ii]
        _ok = ((abs(iths[0] - iths_ref[0]) < 2e-9) and
               nm.isclose(iths[-1], ienergy, atol=0, rtol=e_rtols[kinds[ii]]))
        tst.report(('%d % 20s:' + (6 * ' % .2e'))
                   % ((_ok, kinds[ii]) + tuple(iths)))
        ok = _ok and ok

    assert ok
    assert nm.isclose(e0, 1.8e-4, atol=0, rtol=1e-12)

def test_rmm_solver(problem, output_dir):
    from sfepy.base.base import IndexedStruct

    ls_conf = problem.solver_confs['ls']
    lsr_conf = problem.solver_confs['lsrmm']
    tss_conf = problem.solver_confs['tscd']

    vu = problem.get_variables()['u']
    sensor = problem.domain.regions['Sensor']
    isens = 3 * vu.field.get_dofs_in_region(sensor)[0] + 2

    def store_ths(pb, ts, variables, ths):
        sp = variables.get_state_parts()
        u1, v1 = sp['u'], sp['du']

        e_u = 0.5 * u1 @ pb.Kf @ u1
        e_t = 0.5 * v1 @ pb.Mf @ v1
        ths.append((ts.time, u1[isens], v1[isens], e_u, e_t))

    problem.tsc_conf = None

    status = IndexedStruct()
    problem.init_solvers(ls_conf=ls_conf, ts_conf=tss_conf, status=status,
                         force=True)
    ths = []
    problem.solve(status=status, save_results=False,
                  step_hook=partial(store_ths, ths=ths))
    ths = nm.array(ths)

    statusr = IndexedStruct()
    problem.init_solvers(ls_conf=lsr_conf, ts_conf=tss_conf, status=statusr,
                         force=True)
    thsr = []
    problem.solve(status=statusr, save_results=False,
                  step_hook=partial(store_ths, ths=thsr))
    thsr = nm.array(thsr)

    tst.report(f'solution times: CMM: {status.time}, RMM: {statusr.time}')

    ratio = abs(ths[:,2].max() / ths[:,1].max())

    # import matplotlib.pyplot as plt
    # colors = plt.cm.tab10.colors
    # fig, ax = plt.subplots()
    # ax.plot(ths[:,0], ths[:,3], color=colors[0], ls='-')
    # ax.plot(ths[:,0], ths[:,4], color=colors[1], ls='-')
    # ax.plot(thsr[:,0], thsr[:,3], color=colors[0], ls='--')
    # ax.plot(thsr[:,0], thsr[:,4], color=colors[1], ls='--')
    # fig, ax = plt.subplots()
    # ax.plot(ths[:,0], ratio * ths[:,1], color=colors[0], ls='-')
    # ax.plot(ths[:,0], ths[:,2], color=colors[1], ls='-')
    # ax.plot(thsr[:,0], ratio * thsr[:,1], color=colors[0], ls='--')
    # ax.plot(thsr[:,0], thsr[:,2], color=colors[1], ls='--')
    # plt.show()

    dt = ths[1, 0] - ths[0, 0]
    ierrs = nm.linalg.norm(ths[:, 1:] - thsr[:, 1:], axis=0) * dt

    assert ierrs[0] < 1e-13
    assert ierrs[1] < 3 * ratio * 1e-13
    assert ierrs[2] < 1e-11
    assert ierrs[3] < 1e-11

def test_active_only(output_dir):
    """
    Note: with tsc the results would differ, as eval_scaled_norm() depends on
    the vector length.
    """
    import sys
    from sfepy.discrete import Problem
    from sfepy.base.conf import ProblemConf

    define_dict = define(dims=(0.1, 0.02), shape=(3, 3))
    conf = ProblemConf.from_dict(define_dict, sys.modules[__name__])

    conf.options.active_only = False
    pb = Problem.from_conf(conf)
    pb.tsc_conf = None
    variables_f = pb.solve()

    conf.options.active_only = True
    pb = Problem.from_conf(conf)
    pb.tsc_conf = None
    variables_t = pb.solve()

    assert nm.allclose(variables_f(), variables_t(), atol=1e-7, rtol=0)



#!/usr/bin/env python
"""SfePy: Simple finite elements in Python

SfePy (simple finite elements in Python) is a software, distributed
under the BSD license, for solving systems of coupled partial
differential equations by the finite element method. The code is based
on NumPy and SciPy packages.
"""
import glob
import os

from skbuild import setup
from setuptools import find_packages

import sys
sys.path.append('./tools')
from build_helpers import INFO, cmdclass, log, package_check

from sfepy import config, version

DOCLINES = __doc__.split("\n")

VERSION = INFO.__version__

CLASSIFIERS = """\
Development Status :: 3 - Alpha
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved :: BSD License
Programming Language :: C
Programming Language :: Python
Topic :: Software Development
Topic :: Scientific/Engineering
Operating System :: POSIX
Operating System :: MacOS :: MacOS X
Operating System :: Microsoft :: Windows
"""

DOWNLOAD_URL = "http://sfepy.org/doc-devel/downloads.html"

# BEFORE importing distutils, remove MANIFEST. distutils doesn't properly
# update it when the contents of directories change.
if os.path.exists('MANIFEST'): os.remove('MANIFEST')


def _cython_version(pkg_name):
    from Cython.Compiler.Version import version

    return version


def _igakit_version(pkg_name):
    return '0.1'


def _pymetis_version(pkg_name):
    import pymetis

    return pymetis.version


def _scikit_umfpack_version(pkg_name):
    try:
        import scikits.umfpack; scikits.umfpack
        try:
            return scikits.umfpack.__version__

        except AttributeError:
            return '<0.3.1'

    except:
        return None


def _primme_version(pkg_name):
    import primme

    return '3.0.0'


def check_versions(show_only=False):
    # Cython is a build dependency.
    package_check('cython', INFO.CYTHON_MIN_VERSION,
                  version_getter=_cython_version,
                  show_only=show_only)

    # Check hard and soft dependencies.
    package_check('numpy', INFO.NUMPY_MIN_VERSION,
                  show_only=show_only)
    package_check('scipy', INFO.SCIPY_MIN_VERSION,
                  show_only=show_only)
    package_check('matplotlib', INFO.MATPLOTLIB_MIN_VERSION,
                  show_only=show_only)
    package_check('pyparsing', INFO.PYPARSING_MIN_VERSION,
                  show_only=show_only)
    package_check('tables', INFO.PYTABLES_MIN_VERSION,
                  show_only=show_only)
    package_check('sympy', INFO.SYMPY_MIN_VERSION, optional=True,
                  messages={'opt suffix' : '; some tests are going to fail!'},
                  show_only=show_only)
    package_check('igakit', INFO.IGAKIT_MIN_VERSION, optional=True,
                  version_getter=_igakit_version,
                  show_only=show_only)
    package_check('petsc4py', INFO.PETSC4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('mpi4py', INFO.MPI4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('slepc4py', INFO.SLEPC4PY_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('pymetis', INFO.PYMETIS_MIN_VERSION, optional=True,
                  version_getter=_pymetis_version,
                  show_only=show_only)
    package_check('scikits.umfpack', INFO.SCIKIT_UMFPACK_MIN_VERSION,
                  optional=True,
                  version_getter=_scikit_umfpack_version,
                  show_only=show_only)
    package_check('meshio', INFO.MESHIO_MIN_VERSION,
                  show_only=show_only)
    package_check('psutil', INFO.PSUTIL_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('pyvista', INFO.PYVISTA_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('opt_einsum', INFO.OPT_EINSUM_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('jax', INFO.JAX_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('dask', INFO.DASK_MIN_VERSION, optional=True,
                  show_only=show_only)
    package_check('primme', INFO.PRIMME_MIN_VERSION, optional=True,
                  version_getter=_primme_version, show_only=show_only)


def data_dir_walk(dir_name: str, prefix: str) -> list:
    """
    Generate instructions for setup() to add all files in a tree rooted at `dirname`
    as data_files.
    """
    data_files = []
    for root, dirs, files in os.walk(dir_name):
        full_paths = [os.path.join(root, fname) for fname in files]
        data_files.append((os.path.join(prefix, root), full_paths))

    return data_files


def compose_data_files() -> list:
    data_files = [
        ('sfepy', ['LICENSE', 'VERSION']),
    ]
    test_files = [('sfepy/tests', glob.glob('sfepy/tests/*.py'))]
    mesh_data_files = data_dir_walk('meshes', 'sfepy')
    example_files = data_dir_walk('examples', 'sfepy')

    return data_files + test_files + mesh_data_files + example_files


def cmake_bool(py_bool: bool) -> str:
    return "ON" if py_bool else "OFF"


def compose_cmake_args() -> list:
    conf = config.Config()
    cmake_args = [f'-DCMAKE_C_FLAGS={" ".join(conf.compile_flags())}']

    # Debug flags are always added explicitly, so they won't be taken from cmake cache.
    debug_flags = set(conf.debug_flags())
    cmake_args.append(f"-DDEBUG_FMF={cmake_bool('DEBUG_FMF' in debug_flags)}")
    cmake_args.append(f"-DDEBUG_MESH={cmake_bool('DEBUG_MESH' in debug_flags)}")

    # On Azure CI images, Ninja isn't found automaticvally. Since this is harmless
    # on other platforms, we specify it here.
    # See https://conda-forge.org/docs/maintainer/maintainer_faq.html#mfaq-windows-cmake
    cmake_args.extend(["-G", "Ninja"])

    return cmake_args


def setup_package():
    # Write the version file.
    fd = open('VERSION', 'w')
    fd.write(VERSION)
    fd.close()

    # Create version.h file.
    # There is probably a way to do it with CMake but we'll get to it later.
    filename_in = 'sfepy/discrete/common/extmods/version.h.in'
    filename_out = 'sfepy/discrete/common/extmods/version.h'
    fdi = open(filename_in, 'r')
    fdo = open(filename_out, 'w')
    for line in fdi:
        if line.find('VERSION "0.0.0"') >= 0:
            aux = line.split()
            aux[2] = VERSION
            line = ' '.join(aux) + '\n'
        fdo.write(line)
    fdi.close()
    fdo.close()

    install_requires = [
        'matplotlib',
        'meshio',
        'numpy',
        'pyparsing',
        'pyvista',
        'scipy',
        'sympy',
        'tables',
        'packaging',
    ]

    setup(
        name='sfepy',
        version=version.__version__,
        maintainer="Robert Cimrman",
        maintainer_email="cimrman3@ntc.zcu.cz",
        description=DOCLINES[0],
        long_description="\n".join(DOCLINES[2:]),
        url="http://sfepy.org",
        download_url=DOWNLOAD_URL,
        license='BSD',
        classifiers=list(filter(None, CLASSIFIERS.split('\n'))),
        platforms=["Linux", "Mac OS-X", 'Windows'],
        entry_points={
          'console_scripts': [
              'sfepy-convert=sfepy.scripts.convert_mesh:main',
              'sfepy-mesh=sfepy.scripts.gen_mesh:main',
              'sfepy-probe=sfepy.scripts.probe:main',
              'sfepy-run=sfepy.scripts.simple:main',
              'sfepy-test=sfepy.scripts.run_tests:main',
              'sfepy-view=sfepy.scripts.resview:main',
          ],
        },
        install_requires=install_requires,
        cmdclass=cmdclass,
        packages=find_packages(),
        data_files=compose_data_files(),
        setup_requires=['cython'],
        cmake_args=compose_cmake_args(),
        cmake_languages=('C')
    )


if __name__ == '__main__':
    check_versions()
    setup_package()

    from sfepy import Config
    site_config = Config()
    log.info('\nUsing Python {}.'.format(site_config.python_version()))

    log.info('\nRequired and optional packages found:\n')
    check_versions(show_only=True)